from __future__ import annotations
from typing import List, Type, Union, Dict, Callable, Tuple, TYPE_CHECKING

import os
import sys
import types
import logging
import inspect
import pathlib
import functools
import importlib.util
from _balder.utils import get_class_that_defines_method
from _balder.setup import Setup
from _balder.device import Device
from _balder.feature import Feature
from _balder.vdevice import VDevice
from _balder.scenario import Scenario
from _balder.connection import Connection
from _balder.controllers import ScenarioController, SetupController, DeviceController, VDeviceController, \
    FeatureController, NormalScenarioSetupController
from _balder.exceptions import DuplicateForVDeviceError, UnknownVDeviceException
from _balder.utils import get_scenario_inheritance_list_of

if TYPE_CHECKING:
    from _balder.plugin_manager import PluginManager

logger = logging.getLogger(__file__)


class Collector:
    """
    The Collector class manages the loading and importing of all relevant balder objects. It does not resolve something,
    but secures that all relevant data is being collected.
    """
    # metadata object that contains all raw fixtures (classes that were not be resolved yet)
    raw_fixtures = {}

    # this static attribute will be managed by the decorator `@for_vdevice(..)`. It holds all functions/methods that
    # were decorated with `@for_vdevice(..)` (without checking their correctness). The collector will check them later
    # with the method `rework_method_variation_decorators()`
    _possible_method_variations: Dict[
        Callable,
        List[Tuple[Union[Type[VDevice], str],
                   Union[Type[Connection], Connection, Tuple[Union[Type[Connection], Connection]],
                         List[Union[Type[Connection], Connection, Tuple[Union[Type[Connection], Connection]]]]]]]] = {}

    def __init__(self, working_dir: pathlib.Path):
        self.working_dir = pathlib.Path(working_dir)

        # add the working directory to the official python path
        sys.path.insert(0, str(self.working_dir.parent.absolute()))

        self._all_py_files: Union[List[pathlib.Path], None] = None
        self._all_scenarios: Union[List[Type[Scenario]], None] = None
        self._all_setups: Union[List[Type[Setup]], None] = None
        self._all_connections: Union[List[Type[Connection]], None] = None

        self.balderglob_was_loaded = False

    @property
    def all_pyfiles(self) -> List[pathlib.Path]:
        """returns a list of all python files that were be found by the collector"""
        if self._all_py_files is None:
            raise AttributeError("please call the `collect()` method before omitting this value")
        return self._all_py_files

    @property
    def all_scenarios(self) -> List[Type[Scenario]]:
        """returns a list of all scenarios that were found by the collector"""
        if self._all_scenarios is None:
            raise AttributeError("please call the `collect()` method before omitting this value")
        return self._all_scenarios

    @property
    def all_setups(self) -> List[Type[Setup]]:
        """returns a list of all setups that were found by the collector"""
        if self._all_setups is None:
            raise AttributeError("please call the `collect()` method before omitting this value")
        return self._all_setups

    @property
    def all_scenarios_and_setups(self) -> List[Union[Type[Scenario], Type[Setup]]]:
        """
        returns a list of all scenarios and setups that were found by the collector
        """
        new_list: List[Union[Type[Scenario], Type[Setup]]] = []
        new_list += self.all_scenarios
        new_list += self.all_setups
        return new_list

    @property
    def all_connections(self) -> List[Type[Connection]]:
        """returns a list of all connections that were found by the collector"""
        if self._all_connections is None:
            raise AttributeError("please call the `collect()` method before omitting this value")
        return self._all_connections

    def load_balderglob_py_file(self) -> Union[types.ModuleType, None]:
        """
        This method loads the global balderglob.py file and returns the module or None if the file does not exist.
        """
        filepath = self.working_dir.joinpath('balderglob.py')

        module_name = f"{self.working_dir.stem}.balderglob"

        if self.balderglob_was_loaded:
            return sys.modules[module_name]

        if filepath.is_file():
            self.balderglob_was_loaded = True

            spec = importlib.util.spec_from_file_location(module_name, filepath)
            cur_module = importlib.util.module_from_spec(spec)
            sys.modules[module_name] = cur_module
            spec.loader.exec_module(cur_module)
            return cur_module
        return None

    def get_all_py_files(self) -> List[pathlib.Path]:
        """
        This method returns all python modules that the system can find in the current set WORKING_DIR. It doesn't
        matter what kind of modules it is or what content they offer.
        """
        all_paths: List[pathlib.Path] = []

        for root, _, files in os.walk(str(self.working_dir)):
            for file in files:
                if file.endswith(".py"):
                    all_paths.append(pathlib.Path(os.path.join(root, file)))
        return all_paths

    def get_all_scenario_classes(self, py_file_paths: List[pathlib.Path], filter_abstracts: bool = True) -> \
            List[Type[Scenario]]:
        """
        This method searches all py-file paths that are given with the parameter ``py_file_paths``. It searches for all
        valid balder :class:`Scenario` classes. First the method filters the paths for the files that begin with
        ``scenario_*`` and then searches in the file for classes that begin with ``Scenario*``. Afterwards, it checks
        that the class is a subclass from :class:`Scenario`.

        :param py_file_paths: a list of python files the collector should search through to extract the
                              :meth:`Scenario` classes

        :param filter_abstracts: if this value is True, abstract :meth:`Scenario` classes will be filtered

        :return: a list of all classes (subclasses of :meth:`Scenario`) that are found by the method in the given
                 file paths
        """
        result = []
        for cur_path in py_file_paths:
            #: only use files that match the filter
            if not cur_path.parts[-1].startswith('scenario_'):
                continue
            module_name = \
                f"{self.working_dir.stem}.{'.'.join(cur_path.parent.relative_to(self.working_dir).parts)}." \
                f"{ cur_path.stem}"

            if module_name in sys.modules.keys():
                # ignore all already imported items
                continue

            spec = importlib.util.spec_from_file_location(module_name, cur_path)
            cur_module = importlib.util.module_from_spec(spec)
            sys.modules[module_name] = cur_module
            spec.loader.exec_module(cur_module)
            class_members = inspect.getmembers(cur_module, inspect.isclass)
            for cur_class_name, cur_class in class_members:
                if cur_class_name.startswith('Scenario') and issubclass(cur_class, Scenario) \
                        and cur_class != Scenario:
                    if filter_abstracts and inspect.isabstract(cur_class):
                        break
                    if cur_class not in result:
                        result.append(cur_class)
        return result

    def load_all_connection_classes(self, py_file_paths: List[pathlib.Path]) -> None:
        """
        This method searches all py-file paths that are given with the parameter ``py_file_paths``. It searches for all
        valid balder :class:`Connection` classes. The method imports all classes that are directly located in a
        submodule `connections` (py file or package directory).

        :param py_file_paths: a list of python files that the collector should search through to extract the
                              :meth:`Connection` classes

        :return: a list of all classes (subclasses of :meth:`Connection`) that are found by the method in the
                 given files
        """
        for cur_path in py_file_paths:
            #: only use files that match the filter
            if 'connections' in cur_path.parts[-2] or 'connections.py' == cur_path.parts[-1]:
                module_name = \
                    f"{self.working_dir.stem}.{'.'.join(cur_path.parent.relative_to(self.working_dir).parts)}." \
                    f"{cur_path.stem}"

                spec = importlib.util.spec_from_file_location(module_name, cur_path)
                cur_module = importlib.util.module_from_spec(spec)
                sys.modules[module_name] = cur_module
                spec.loader.exec_module(cur_module)

    def get_all_connection_classes(self) -> List[Type[Connection]]:
        """
        The method determines all available classes that are inherited from the base class :meth:`Connection`.

        :return: a list of all classes (subclasses of :meth:`Connection`) that are found in all imported modules
        """
        result = []
        modules = sys.modules.copy()
        for _, cur_module in modules.items():
            try:
                for _, cur_class in inspect.getmembers(cur_module, inspect.isclass):
                    if issubclass(cur_class, Connection) and cur_class not in result:
                        result.append(cur_class)
            except ModuleNotFoundError:
                # skip these errors if some modules has wrong module references
                pass
        return result

    def get_all_setup_classes(self, py_file_paths: List[pathlib.Path], filter_abstracts: bool = True) -> \
            List[Type[Setup]]:
        """
        This method searches all py-file paths that are given with the parameter ``py_file_paths``. It searches for all
        valid balder :class:`Setup` classes. First the method filters the paths for the files that begin with
        ``setup_*`` and then searches in the file for classes that begin with ``Setup*``. Afterwards, it checks
        that the class is a subclass from :class:`Setup`.

        :param py_file_paths: a list of python files the collector should search through to extract the
                              :meth:`Setup` classes

        :param filter_abstracts: if this value is True, abstract :meth:`Setup` classes will be filtered

        :return: a list of all classes (subclasses of :class:`Setup`) that are found by the method in the given
                 file paths
        """
        result = []
        for cur_path in py_file_paths:
            #: only use files that match the filter
            if not cur_path.parts[-1].startswith('setup_'):
                continue
            module_name = \
                f"{self.working_dir.stem}.{'.'.join(cur_path.parent.relative_to(self.working_dir).parts)}." \
                f"{cur_path.stem}"

            # ignore all already imported items
            if module_name in sys.modules.keys():
                continue

            spec = importlib.util.spec_from_file_location(module_name, cur_path)
            cur_module = importlib.util.module_from_spec(spec)
            sys.modules[module_name] = cur_module
            spec.loader.exec_module(cur_module)
            class_members = inspect.getmembers(cur_module, inspect.isclass)
            for cur_class_name, cur_class in class_members:
                if cur_class_name.startswith('Setup') and issubclass(cur_class, Setup) and cur_class != Setup:
                    if filter_abstracts and inspect.isabstract(cur_class):
                        break
                    if cur_class not in result:
                        result.append(cur_class)
        return result

    @staticmethod
    def filter_parent_classes_of(items: List[Union[Type[Scenario], Type[Setup]]]):
        """
        This method removes all parent classes from the given items. All items that also have children, will be removed
        from the returned list.
        """
        result = []
        for cur_class in items:
            a_child_class_exists = False
            for cur_other_class in items:
                if cur_class == cur_other_class:
                    # the current class is the class itself
                    continue
                if issubclass(cur_other_class, cur_class):
                    a_child_class_exists = True
                    break
            if not a_child_class_exists:
                result.append(cur_class)
        return result

    def set_run_skip_ignore_of_test_methods_in_scenarios(self):
        """
        This method secures that the scenario classes have a valid RUN, SKIP and IGNORE attribute.
        """
        for cur_scenario in self._all_scenarios:
            # determines hierarchy of inherited classes
            base_classes = get_scenario_inheritance_list_of(cur_scenario)
            # removes the last scenario class
            base_classes = base_classes[:-1]
            # now determine all RUN, SKIP and IGNORE values if they aren't already mentioned - if there exists a value
            #  for them, check if the value is valid
            base_classes.reverse()
            for cur_idx, cur_class in enumerate(base_classes):
                next_parent = None if cur_idx == 0 else base_classes[cur_idx - 1]
                if "IGNORE" in cur_class.__dict__.keys():
                    # IGNORE is mentioned in this specific class
                    if not isinstance(cur_class.IGNORE, list):
                        raise TypeError(f"the class attribute `{cur_class.__name__}.IGNORE` has to be from type list")
                    # check that all elements that are mentioned here exists in the parent's RUN, SKIP or IGNORE class
                    # variable or are defined in this class
                    for cur_ignore_method in cur_class.IGNORE:
                        if not inspect.ismethod(cur_ignore_method) or \
                                not cur_ignore_method.__name__.startswith('test_'):
                            raise TypeError(f"the given element {cur_ignore_method.__name__} for class attribute "
                                            f"`{cur_class.__name__}.IGNORE` is no valid test method")
                        if cur_ignore_method not in cur_class.__dict__.values() and next_parent is None or \
                                cur_ignore_method not in cur_class.__dict__.values() and next_parent is not None \
                                and cur_ignore_method not in next_parent.RUN + next_parent.SKIP + next_parent.IGNORE:
                            raise ValueError(f"the element `{cur_ignore_method.__name__}` given at class attribute "
                                             f"`{cur_class.__name__}.IGNORE` is not a test method of this scenario")
                else:
                    # IGNORE is not mentioned in this specific class -> so add an empty list for further access
                    cur_class.IGNORE = []
                if "SKIP" in cur_class.__dict__.keys():
                    # SKIP is mentioned in this specific class
                    if not isinstance(cur_class.SKIP, list):
                        raise TypeError(f"the class attribute `{cur_class.__name__}.SKIP` has to be from type list")
                    # check that all elements that are mentioned here exist in the parent's RUN or SKIP class variable
                    # or are defined in this class
                    for cur_skip_method in cur_class.SKIP:
                        if not inspect.ismethod(cur_skip_method) or \
                                not cur_skip_method.__name__.startswith('test_'):
                            raise TypeError(f"the given element {cur_skip_method.__name__} for class attribute "
                                            f"`{cur_class.__name__}.IGNORE` is no valid test method")
                        if cur_skip_method not in cur_class.__dict__.values():
                            if next_parent is None:
                                raise ValueError(f"the element `{cur_skip_method.__name__}` given at class attribute "
                                                 f"`{cur_class.__name__}.SKIP` is not a test method of this scenario")

                            if cur_skip_method in next_parent.IGNORE:
                                raise ValueError(f"the element `{cur_skip_method.__name__}` given at class "
                                                 f"attribute `{cur_class.__name__}.SKIP` was already added to "
                                                 f"IGNORE in a higher parent class - not possible to add it now to "
                                                 f"SKIP")

                            if cur_skip_method not in next_parent.SKIP and cur_skip_method not in next_parent.IGNORE:
                                raise ValueError(f"the element `{cur_skip_method.__name__}` given at class "
                                                 f"attribute `{cur_class.__name__}.SKIP` is not a test method of "
                                                 f"this scenario")
                else:
                    # SKIP is not mentioned in this specific class -> so add an empty list for further access
                    cur_class.SKIP = []
                if "RUN" in cur_class.__dict__.keys():
                    # RUN is mentioned in this specific class
                    if not isinstance(cur_class.RUN, list):
                        raise TypeError(f"the class attribute `{cur_class.__name__}.RUN` has to be from type list")
                    # check that all elements that are mentioned here exists in the parent's RUN class variable or are
                    # defined in this class
                    for cur_run_method in cur_class.RUN:
                        if not inspect.ismethod(cur_run_method) or \
                                not cur_run_method.__name__.startswith('test_'):
                            raise TypeError(f"the given element {cur_run_method.__name__} for class attribute "
                                            f"`{cur_class.__name__}.RUN` is no valid test method")
                        if cur_run_method not in cur_class.__dict__.values():
                            if next_parent is None:
                                raise ValueError(
                                    f"the element `{cur_run_method.__name__}` given at class attribute "
                                    f"`{cur_class.__name__}.RUN` is not a test method of this scenario")

                            if cur_run_method in next_parent.IGNORE:
                                raise ValueError(
                                    f"the element `{cur_run_method.__name__}` given at class attribute "
                                    f"`{cur_class.__name__}.RUN` was already added to IGNORE in a higher parent "
                                    f"class - not possible to add it now to RUN")
                            if cur_run_method in next_parent.SKIP:
                                raise ValueError(
                                    f"the element `{cur_run_method.__name__}` given at class attribute "
                                    f"`{cur_class.__name__}.RUN` was already added to SKIP in a higher parent "
                                    f"class - not possible to add it now to RUN")
                            if cur_run_method not in next_parent.RUN:
                                raise ValueError(
                                    f"the element `{cur_run_method.__name__}` given at class attribute "
                                    f"`{cur_class.__name__}.RUN` is not a test method of this scenario")
                else:
                    # RUN is not mentioned in this specific class -> so add an empty list for further access
                    cur_class.RUN = []

                # also add all in this class defined test methods to this RUN list
                for cur_item_name, cur_item in cur_class.__dict__.items():
                    if cur_item_name.startswith("test_") and inspect.ismethod(cur_item):
                        cur_class.RUN.append(cur_item)
                # now add all items from parent classes to the lists that are not mentioned yet
                if next_parent is not None:
                    for cur_parent_run_method in next_parent.RUN:
                        if cur_parent_run_method not in cur_class.RUN and cur_parent_run_method not in cur_class.SKIP \
                                and cur_parent_run_method not in cur_class.IGNORE:
                            cur_class.RUN.append(cur_parent_run_method)
                    for cur_parent_skip_method in next_parent.SKIP:
                        if cur_parent_skip_method not in cur_class.SKIP \
                                and cur_parent_skip_method not in cur_class.IGNORE:
                            cur_class.SKIP.append(cur_parent_skip_method)
                    for cur_parent_ignore_method in next_parent.IGNORE:
                        if cur_parent_ignore_method not in cur_class.IGNORE:
                            cur_class.IGNORE.append(cur_parent_ignore_method)

    @staticmethod
    def rework_method_variation_decorators():
        """
        This method iterates over the static attribute `Collector._possible_method_variations` and checks if these
        decorated functions are valid (if they are methods of a :meth:`Feature` class). All valid decorated data will
        then be set for the related feature classes.
        """

        for cur_fn, cur_decorator_data_list in Collector._possible_method_variations.items():
            owner = get_class_that_defines_method(cur_fn)
            owner_feature_controller = FeatureController.get_for(owner)
            name = cur_fn.__name__

            owner_for_vdevice = owner_feature_controller.get_method_based_for_vdevice()
            owner_for_vdevice = owner_for_vdevice if owner_for_vdevice is not None else {}

            for cur_decorator_vdevice, cur_decorator_with_connections in cur_decorator_data_list:

                if owner is None:
                    raise TypeError("The use of the `@for_vdevice` decorator is only allowed for `Feature` objects and "
                                    "methods of `Feature` objects")
                if not issubclass(owner, Feature):
                    raise TypeError(f"The use of the `for_vdevice` decorator is not allowed for methods of a "
                                    f"`{owner.__name__}`. You should only use this decorator for `Feature` elements")

                if isinstance(cur_decorator_vdevice, str):
                    # vDevice is a string, so we have to convert it to the correct class
                    cur_decorator_vdevice = \
                        owner_feature_controller.get_inner_vdevice_class_by_string(cur_decorator_vdevice)

                    if cur_decorator_vdevice is None:
                        raise ValueError(f"can not find a matching inner VDevice class for the given vdevice string "
                                         f"`{cur_decorator_vdevice}` in the feature class `{owner.__name__}`")

                if cur_decorator_vdevice not in owner_feature_controller.get_abs_inner_vdevice_classes():
                    raise UnknownVDeviceException(f"the given vDevice `{cur_decorator_vdevice.__name__}` is not a "
                                                  f"usable vDevice in Feature class `{owner.__name__}`")

                if name not in owner_for_vdevice.keys():
                    owner_for_vdevice[name] = {}

                cur_decorator_cleaned_cnns = []
                for cur_cnn in cur_decorator_with_connections:
                    cur_cnn = cur_cnn() if isinstance(cur_cnn, type) and issubclass(cur_cnn, Connection) else cur_cnn
                    cur_decorator_cleaned_cnns.append(cur_cnn)

                if cur_fn in owner_for_vdevice[name].keys():
                    old_dict = owner_for_vdevice[name][cur_fn]
                    if cur_decorator_vdevice in old_dict.keys():
                        raise DuplicateForVDeviceError(f'there already exists a decorator for the vDevice '
                                                       f'`{cur_decorator_vdevice}` at method `{name}` of class '
                                                       f'`{owner.__name__}` ')

                    old_dict[cur_decorator_vdevice] = cur_decorator_cleaned_cnns
                    owner_for_vdevice[name][cur_fn] = old_dict
                else:
                    new_dict = {cur_decorator_vdevice: cur_decorator_cleaned_cnns}
                    owner_for_vdevice[name][cur_fn] = new_dict

            def owner_wrapper(the_owner_of_this_method, the_name, wrap_fn):
                @functools.wraps(wrap_fn)
                def method_variation_multiplexer(this, *args, **kwargs):
                    controller = FeatureController.get_for(this.__class__)
                    if this.__class__ == the_owner_of_this_method:
                        # this is no parent class call -> use the set method-variation (set by VariationExecutor before)
                        _, _, func = controller.get_active_method_variation(the_name)
                    else:
                        func = controller.get_inherited_method_variation(the_owner_of_this_method, the_name)
                        if func is None:
                            raise AttributeError(f'there exists no method/method variation in class '
                                                 f'`{the_owner_of_this_method.__name__}` or its parent classes')
                    if len(args) == 0:
                        return func(self=this, **kwargs)

                    new_args = [this, *args]
                    return func(*new_args)

                return method_variation_multiplexer

            new_callback = owner_wrapper(owner, name, cur_fn)
            setattr(owner, name, new_callback)
            owner_feature_controller.set_method_based_for_vdevice(owner_for_vdevice)

    def get_all_scenario_feature_classes(self) -> List[Type[Feature]]:
        """
        This method returns a list with all :class:`Feature` classes that are being instantiated in one or more
        :class:`Scenario` classes.

        :return: a list with all :class:`Feature` classes
        """
        all_scenario_features = []
        for cur_scenario in self._all_scenarios:
            for cur_device in ScenarioController.get_for(cur_scenario).get_all_abs_inner_device_classes():
                all_instantiated_features = DeviceController.get_for(cur_device).get_all_instantiated_feature_objects()
                for _, cur_feature_obj in all_instantiated_features.items():
                    if cur_feature_obj.__class__ not in all_scenario_features:
                        all_scenario_features.append(cur_feature_obj.__class__)
        return all_scenario_features

    def get_all_setup_feature_classes(self) -> List[Type[Feature]]:
        """
        This method returns a list with all :class:`Feature` classes that are instantiated in :class:`Setup` class
        devices.

        :return: a list with all :class:`Feature` classes
        """
        all_setup_features = []
        for cur_setup in self._all_setups:
            for cur_device in SetupController.get_for(cur_setup).get_all_abs_inner_device_classes():
                cur_device_instantiated_features = \
                    DeviceController.get_for(cur_device).get_all_instantiated_feature_objects()
                for _, cur_feature_obj in cur_device_instantiated_features.items():
                    if cur_feature_obj.__class__ not in all_setup_features:
                        all_setup_features.append(cur_feature_obj.__class__)
        return all_setup_features

    def get_all_scenario_device_classes(self) -> List[Type[Device]]:
        """
        This method returns a list with all inner :class:`Device` classes that are defined in one of the collected
        :class:`.Scenario` classes.

        :return: a list with all :class:`Device` classes
        """
        all_scenario_devices = []
        for cur_scenarios in self._all_scenarios:
            all_scenario_devices += ScenarioController.get_for(cur_scenarios).get_all_abs_inner_device_classes()
        return all_scenario_devices

    def get_all_setup_device_classes(self) -> List[Type[Device]]:
        """
        This method returns a list with all inner :class:`Device` classes that are defined in one of the collected
        :class:`.Setup` classes.

        :return: a list with all :class:`Device` classes
        """
        all_setup_devices = []
        for cur_setup in self._all_setups:
            all_setup_devices += SetupController.get_for(cur_setup).get_all_abs_inner_device_classes()
        return all_setup_devices

    def _set_original_vdevice_in_features(self):
        """
        This method ensures that the original VDevice classes (that are defined in the :class:`Feature` classes) are
        saved inside their controllers.
        """
        for cur_feature in self.get_all_scenario_feature_classes() + self.get_all_setup_feature_classes():
            cur_feature_controller = FeatureController.get_for(cur_feature)
            cur_feature_controller.save_all_current_vdevice_references_as_originals()

    def _set_original_device_features(self):
        """
        This method ensures that the original features (that are instantiated in the
        :class:`Scenario`/:class:`Setup` devices or in the :class:`VDevice`) are saved inside their controllers.
        """

        for cur_scenario_or_setup in self.all_scenarios_and_setups:
            cur_scenario_or_setup_controller = NormalScenarioSetupController.get_for(cur_scenario_or_setup)
            devices = cur_scenario_or_setup_controller.get_all_abs_inner_device_classes()
            for cur_device in devices:
                cur_device_controller = DeviceController.get_for(cur_device)
                cur_device_controller.save_all_original_instanced_features()

        for cur_feature in self.get_all_scenario_feature_classes() + self.get_all_setup_feature_classes():
            vdevices = FeatureController.get_for(cur_feature).get_abs_inner_vdevice_classes()
            for cur_vdevice in vdevices:
                cur_vdevice_controller = VDeviceController.get_for(cur_vdevice)
                cur_vdevice_controller.save_all_original_instanced_features()

    def _exchange_strings_with_objects(self):
        """
        This method exchanges all strings (that can be used in decorators) are exchanged with their real objects. It
        secures this for all :class:`Device` and :class:`VDevice` references inside the session.
        """
        # resolve connection Device strings (for all devices that are directly defined inside the scenario/setup)
        for cur_scenario_or_setup in self.all_scenarios_and_setups:
            cur_scenario_or_setup_controller = NormalScenarioSetupController.get_for(cur_scenario_or_setup)
            for cur_device in cur_scenario_or_setup_controller.get_all_inner_device_classes():
                DeviceController.get_for(cur_device).resolve_connection_device_strings()

        # resolve connection VDevice strings (for all absolute devices of this scenario/setup)
        for cur_scenario_or_setup in self.all_scenarios_and_setups:
            cur_scenario_or_setup_controller = NormalScenarioSetupController.get_for(cur_scenario_or_setup)
            for cur_device in cur_scenario_or_setup_controller.get_all_abs_inner_device_classes():
                DeviceController.get_for(cur_device).resolve_mapped_vdevice_strings()

    def _validate_scenario_and_setups(self):
        """
        This method validates that the inheritance of `Setup`/`Scenario` classes were being done correctly. It checks
        that all devices that are inherited has the same naming as their parents and also that every reused name (that
        is already be used for a device in parent class) does also inherit from the parent scenario/setup device.

        In addition to that, the method checks that either every device of higher class is defined (and overwritten) in
        the current class or non device
        """
        for cur_scenario_or_setup in self.all_scenarios_and_setups:
            cur_scenario_or_setup_controller = NormalScenarioSetupController.get_for(cur_scenario_or_setup)
            cur_scenario_or_setup_controller.validate_inheritance()

    def _validate_features_and_their_vdevices(self):
        """
        This method validates the correct usage of features and their vDevices. It secures that the feature inheritance
        inside devices is correct and validates that inner referenced features inside other feature are correct. In
        addition to that it secures that inner vDevice classes of features are inherited and are used correctly.
        """
        all_features = self.get_all_scenario_feature_classes() + self.get_all_setup_feature_classes()
        all_devices = self.get_all_scenario_device_classes() + self.get_all_setup_device_classes()

        # validate inheritance for all features
        for cur_device in all_devices:
            DeviceController.get_for(cur_device).validate_inheritance_of_instantiated_features()

        # validate inner references of features inside features
        for cur_outer_device in all_devices:
            DeviceController.get_for(cur_outer_device).validate_inner_referenced_features()

        # validate all inner classes of all features and secure that Device subclasses are correctly used
        for cur_feature in all_features:
            FeatureController.get_for(cur_feature).validate_inner_classes()

        # validate that all parent VDevices are overwritten if one or more VDevice(s) are defined in current feature
        for cur_feature in all_features:
            cur_feature_controller = FeatureController.get_for(cur_feature)
            cur_feature_controller.validate_inner_vdevice_inheritance()

    def _determine_class_based_values_for_all_features(self, print_warning=True):
        """
        This method determines the absolute class based values for `@for_vdevice`. It will not update any internal
        values if the class-based-value was already set by an explicit class based `@for_vdevice` decorator. Otherwise,
        it determines the class based `@for_vdevice` value through analysing the method-based-decorators and sets this
        determined value. If the method finds a better value, it throws a warning with a suggestion for the (nicer)
        class based decorator.

        In both cases the method will secure that every given vDevice class is a real part of the current
        :class:`Feature` class (will be returned by direct call of method `Feature.get_inner_vdevice_classes()`).


        .. note::
            This method automatically updates the values for the parent classes, too. Every time it searches for the
            values it also considers the parent values for the vDevice or the parent class of the vDevice.
        """
        for cur_feature in self.get_all_scenario_feature_classes() + self.get_all_setup_feature_classes():
            cur_feature_controller = FeatureController.get_for(cur_feature)
            cur_feature_controller.determine_absolute_class_based_for_vdevice(print_warning)

    def _validate_vdevice_feature_references(self):
        """
        This method validates that VDevice references are used correctly.
        """
        # check feature (referenced from VDevice) exists in the related VDevice
        for cur_scenario_or_setup in self.all_scenarios_and_setups:
            cur_scenario_or_setup_controller = NormalScenarioSetupController.get_for(cur_scenario_or_setup)
            cur_scenario_or_setup_controller.check_vdevice_feature_existence()

        # secure for all scenario features that the class based for_vdevice values of a child :class:`Feature` class
        # are contained_in the related VDevice defined in a parent :class:`Feature` class.
        for cur_feature in self.get_all_scenario_feature_classes():
            FeatureController.get_for(cur_feature).validate_inherited_class_based_vdevice_cnn_subset()

        # validates for every active class-based feature (only the ones that have an active VDevice<->Device mapping),
        # that a clear scenario-device-connection exists for this feature
        for cur_scenario in self.all_scenarios:
            ScenarioController.get_for(cur_scenario).validate_feature_clearance_for_parallel_connections()

    def _determine_absolute_device_connections(self):
        """
        This method manages the absolute connections for all collected :class:`Scenario` and :class:`Setup` objects. It
        determines the raw absolute connections for scenarios and setups and also creates the real device connections
        for all scenarios.
        """
        # determine the raw absolute device connections for all scenarios and setups
        for cur_scenario_or_setup in self.all_scenarios_and_setups:
            NormalScenarioSetupController.get_for(cur_scenario_or_setup).determine_raw_absolute_device_connections()

        # determine all absolute device connections (only for scenarios)
        for cur_scenario in self.all_scenarios:
            ScenarioController.get_for(cur_scenario).determine_absolute_device_connections()

    def _validate_feature_connections_in_setup(self):
        """
        This method validates that the connections of setup features are valid.
        It ensures, that every feature connection (that already has a vDevice<->Device mapping on setup level)
        has a connection that is CONTAINED-IN the connection between the related setup devices.
        """
        for cur_setup in self.all_setups:
            SetupController.get_for(cur_setup).validate_feature_possibility()

    def collect(self, plugin_manager: PluginManager):
        """
        This method manages the entire collection process.

        :param plugin_manager: contains the reference to the used plugin manager
        """
        # load all py files
        self.load_balderglob_py_file()
        self._all_py_files = self.get_all_py_files()
        self._all_py_files = plugin_manager.execute_modify_collected_pyfiles(self._all_py_files)

        # collect all `Connection` classes (has to be first, because scenarios/setups can use them)
        self.load_all_connection_classes(py_file_paths=self._all_py_files)
        self._all_connections = self.get_all_connection_classes()

        # collect all `Scenario` classes
        self._all_scenarios = self.get_all_scenario_classes(
            py_file_paths=self._all_py_files, filter_abstracts=True)
        # collect all `Setup` classes
        self._all_setups = self.get_all_setup_classes(
            py_file_paths=self._all_py_files, filter_abstracts=True)

        self._all_scenarios, self._all_setups = plugin_manager.execute_collected_classes(
            scenarios=self._all_scenarios, setups=self._all_setups)

        self._all_scenarios = Collector.filter_parent_classes_of(items=self._all_scenarios)
        self._all_setups = Collector.filter_parent_classes_of(items=self._all_setups)

        Collector.rework_method_variation_decorators()

        # do some further stuff after everything was read
        self._set_original_vdevice_in_features()
        self._set_original_device_features()
        self._exchange_strings_with_objects()

        # todo move this implementation in related controller
        self.set_run_skip_ignore_of_test_methods_in_scenarios()

        self._validate_scenario_and_setups()

        self._validate_features_and_their_vdevices()

        self._determine_class_based_values_for_all_features()

        self._validate_vdevice_feature_references()

        self._determine_absolute_device_connections()

        self._validate_feature_connections_in_setup()



from multiprocessing import Process

from balder.exceptions import ConnectionIntersectionError
from _balder.balder_session import BalderSession


def test_1_scenario_illegal_vdevice_cnn(balder_working_dir):
    """
    This testcase executes the basic CALCULATOR example. It maps a device to a vDevice, which requires a connection that
    does not exist in the scenario class itself. The VDevice<->Device mapping is also done in the scenario.

    The test expects the throwing of the ``ConnectionIntersectionError`` on COLLECTOR level.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    try:
        session.run()
        print("\n")
        assert False, "test session terminates without an error"
    except ConnectionIntersectionError as exc:
        assert exc.args[0] == "the `ScenarioAdding` has a connection from device `NumberOneDevice` to `Calculator` - " \
                              "some mapped VDevices of their feature classes define mismatched connections"

    assert session.executor_tree is None, "test session does not terminates before collector work was done"



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_1_devicemethod(balder_working_dir):
    """
    This testcase validates an environment with a Scenario-Device that has an own method. It will be checked if the
    method code uses the assigned Setup-Features correctly.

    The test tries a class method and a static method. It fails if one of the methods raises a NotImplementedError.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    session.run()
    print("\n")

    assert session.executor_tree.executor_result == ResultState.SUCCESS, \
        "test session does not terminates with success"



import balder
import logging
from ..lib.features import FeatureI, FeatureII
from ..lib.connections import AConnection
from ..balderglob import RuntimeObserver
from .scenario_a import ScenarioA

logger = logging.getLogger(__file__)


class ScenarioAChild(ScenarioA):
    """This is the CHILD scenario of category A"""

    class ScenarioDevice1(ScenarioA.ScenarioDevice1):
        i = FeatureI()

    @balder.connect(ScenarioDevice1, over_connection=AConnection)
    class ScenarioDevice2(ScenarioA.ScenarioDevice2):
        ii = FeatureII()

    def test_a_1(self):
        RuntimeObserver.add_entry(__file__, ScenarioAChild, ScenarioAChild.test_a_1, category="testcase",
                                  msg=f"execute Test `{ScenarioAChild.test_a_1.__qualname__}`")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

    def test_a_2(self):
        RuntimeObserver.add_entry(__file__, ScenarioAChild, ScenarioAChild.test_a_2, category="testcase",
                                  msg=f"execute Test `{ScenarioAChild.test_a_2.__qualname__}`")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()



from multiprocessing import Process
from _balder.balder_session import BalderSession


def test_collect_only_from_namespace(balder_working_dir):
    """
    This test checks if balder only collects the relevant classes from the correct naming convention files.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n")
    session = BalderSession(cmd_args=["--collect-only"], working_dir=env_dir)
    session.run()
    print("\n")
    from .env.scenarios.scenario_file import ScenarioMatching
    from .env.scenarios.no_scenario_file import ScenarioNotMatching

    all_collected_scenario_names = [cur_scenario.__name__ for cur_scenario in session.all_collected_scenarios]
    all_collected_setup_names = [cur_setup.__name__ for cur_setup in session.all_collected_setups]

    assert ScenarioMatching.__name__ in all_collected_scenario_names, \
        f"can not find the scenario `{ScenarioMatching.__name__}` in collected data"
    assert ScenarioNotMatching.__name__ not in all_collected_scenario_names, \
        f"find the scenario `{ScenarioNotMatching.__name__}` in collected data, where the scenario file name doesn't " \
        f"match the requirements"

    from .env.setups.setup_file import SetupMatching
    from .env.setups.no_setup_file import SetupNoMatching

    assert SetupMatching.__name__ in all_collected_setup_names, \
        f"can not find the setup `{SetupMatching.__name__}` in collected data"
    assert SetupNoMatching.__name__ not in all_collected_setup_names, \
        f"find the setup `{SetupNoMatching.__name__}` in collected data, where the setup file name doesn't match the " \
        f"requirements"



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_1_no_vdevice_mapping(balder_working_dir):
    """
    This testcase executes the basic example and checks if the tree ends in the state NOT_RUN. All variations should not
    be executed, because there is a different vDevice mapping in scenario and in setup.

    The test is the expanded version of the basic calculator example. Its feature `ProvidesANumberFeature` has an
    additional VDevice that is not mapped here in this version.

    The test expects, that balder terminates without executing one test.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    session.run()
    print("\n")

    assert session.executor_tree.executor_result == ResultState.NOT_RUN, \
        "test session does not terminates with error"
    assert len(session.executor_tree.all_child_executors) == 0, "found not expected child elements in the tree"





class FixtureReturn:
    """helper const class for return values"""
    BALDERGLOB_SESSION = "balderglob_session_fixt"
    BALDERGLOB_SETUP = "balderglob_setup_fixt"
    BALDERGLOB_SCENARIO = "balderglob_scenario_fixt"
    BALDERGLOB_VARIATION = "balderglob_variation_fixt"
    BALDERGLOB_TESTCASE = "balderglob_testcase_fixt"

    SETUP_SESSION = "setup_session_fixt"
    SETUP_SETUP = "setup_setup_fixt"
    SETUP_SCENARIO = "setup_scenario_fixt"
    SETUP_VARIATION = "setup_variation_fixt"
    SETUP_TESTCASE = "setup_testcase_fixt"

    SCENARIO_SESSION = "scenario_session_fixt"
    SCENARIO_SETUP = "scenario_setup_fixt"
    SCENARIO_SCENARIO = "scenario_scenario_fixt"
    SCENARIO_VARIATION = "scenario_variation_fixt"
    SCENARIO_TESTCASE = "scenario_testcase_fixt"




from __future__ import annotations
from typing import List, Union, TYPE_CHECKING

import sys
import time
import traceback
from _balder.utils import inspect_method
from _balder.testresult import ResultState, TestcaseResult
from _balder.executor.basic_executor import BasicExecutor
from _balder.previous_executor_mark import PreviousExecutorMark

if TYPE_CHECKING:
    from _balder.executor.variation_executor import VariationExecutor
    from _balder.fixture_manager import FixtureManager
    from _balder.scenario import Scenario


class TestcaseExecutor(BasicExecutor):
    """
    A TestcaseExecutor class represents an actual single test that can be executed. It therefore references exactly to a
    test method of a scenario that can be executed on the specific setup this executor belongs to.
    """

    def __init__(self, testcase: callable, parent: VariationExecutor):
        super().__init__()

        self._base_testcase_callable = testcase
        self._parent_executor = parent
        self._fixture_manager: FixtureManager = parent.fixture_manager

        # if there exist executors that cover this item, they are contained as list in this property
        self.covered_by_executors = None

        # contains the result object for the BODY part of this branch
        self.body_result = TestcaseResult(self)

        # holds the raw test execution time in seconds
        self.test_execution_time_sec = 0

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS ----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def all_child_executors(self) -> None:
        return None

    @property
    def parent_executor(self) -> VariationExecutor:
        return self._parent_executor

    @property
    def base_instance(self) -> object:
        """
        returns the base class instance to which this executor instance belongs"""
        return self._base_testcase_callable

    @property
    def base_testcase_callable(self) -> callable:
        """returns the testcase function"""
        return self._base_testcase_callable

    @property
    def base_testcase_obj(self) -> Scenario:
        """
        return the testcase Scenario this testcase is defined in
        """
        return self._parent_executor.cur_scenario_class

    @property
    def fixture_manager(self):
        """returns the current active fixture manager"""
        return self._fixture_manager

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    def _prepare_execution(self):
        print(f"      TEST {self.base_testcase_callable.__qualname__} ", end='')
        if self.should_be_skipped():
            self.body_result.set_result(ResultState.SKIP)
            self.execution_time_sec = 0
            print("[S]")
            return

    def _body_execution(self):
        start_time = time.perf_counter()
        try:
            _, func_type = inspect_method(self.base_testcase_callable)
            all_args = self.fixture_manager.get_all_attribute_values(
                self, self.base_testcase_obj.__class__, self.base_testcase_callable, func_type)
            if func_type == "staticmethod":
                # testcase is a staticmethod - no special first attribute
                self.base_testcase_callable(**all_args)
            elif func_type == "classmethod":
                self.base_testcase_callable(self=self.base_testcase_obj.__class__, **all_args)
            elif func_type == "instancemethod":
                self.base_testcase_callable(self=self.base_testcase_obj, **all_args)
            else:
                # `function` is not allowed here!
                raise ValueError(f"found illegal value for func_type `{func_type}` for test "
                                 f"`{self.base_testcase_callable.__name__}`")

            self.body_result.set_result(ResultState.SUCCESS)
        except Exception as exc:
            # this has to be a test error
            traceback.print_exception(*sys.exc_info())
            self.body_result.set_result(ResultState.FAILURE, exc)
        self.test_execution_time_sec = time.perf_counter() - start_time

    def _cleanup_execution(self):
        print(f"[{self.body_result.get_result_as_char()}]")

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def has_runnable_elements(self) -> bool:
        return self.prev_mark == PreviousExecutorMark.RUNNABLE

    def should_run(self):
        """returns true if the testcase should be executed (defined in scenario)"""
        if self.base_testcase_callable in self.parent_executor.parent_executor.all_run_tests:
            return True
        return False

    def should_be_skipped(self):
        """returns true if the testcase should be skipped (defined in scenario)"""
        if self.base_testcase_callable in self.parent_executor.parent_executor.all_skip_tests:
            return True
        return False

    def should_be_ignored(self):
        """returns true if the testcase should be ignored (defined in scenario)"""
        if self.base_testcase_callable in self.parent_executor.parent_executor.all_ignore_tests:
            return True
        return False

    def cleanup_empty_executor_branches(self):
        """
        This method searches the whole tree and removes branches where an executor item has no own children. It can
        remove these branches, because they have no valid matchings.

        This method implementation of the :class:`TestcaseExecutor` does nothing.
        """

    def get_covered_by_element(self) -> List[Union[Scenario, callable]]:
        """
        This method returns a list of elements where the whole scenario is covered from. This means, that the whole
        test methods in this scenario are already be covered from every single element in the list.
        """
        all_covered_by_data = []
        scenario_executor = self.parent_executor.parent_executor
        scenario_class = scenario_executor.base_scenario_class
        covered_by_dict_resolved = scenario_executor.get_covered_by_dict()
        if self.base_testcase_callable in covered_by_dict_resolved.keys():
            all_covered_by_data += covered_by_dict_resolved[self.base_testcase_callable]
        if scenario_class in covered_by_dict_resolved.keys():
            all_covered_by_data += covered_by_dict_resolved[scenario_class]
        return all_covered_by_data



from __future__ import annotations

from _balder.device import Device


class VDevice(Device):
    """
    This is the main class for vDevices that will be used as inner classes in :class:`Feature` classes.
    """



from multiprocessing import Process

from _balder.balder_session import BalderSession


def test_1_duplicate_vdevice_mapping(balder_working_dir):
    """
    This testcase executes the basic example and checks that the collector throws an error and the tree will not be
    created.

    The test is the expanded version of the basic calculator example. Its feature `ProvidesANumberFeature` has two
    additional VDevice classes that are both mapped here in this version (in same constructor). This is not allowed.
    This test expects that balder will throw an exception for this.
    To map another vDevice this test devices an additional not connected device `DoNothingDevice`.

    The test expects, that balder throws an error, because the VDevice mapping is not done correctly.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    try:
        session.run()
        print("\n")
        assert False, "the expected `AttributeError` exception was not being thrown"
    except AttributeError as exc:
        assert exc.args[0] == "the constructor expects exactly none or one vDevice mapping - found more than one here"

    assert session.executor_tree is None, "test session does not terminates before collector work was done"



from _balder.testresult import ResultState
from _balder.balder_session import BalderSession

from tests.test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0StaticmethodFixture(Base0EnvtesterClass):
    """
    This testcase executes the envtester environment. This is a reduced version of the environment tester. The test
    checks if fixtures that are defined as static methods (only in scenario and setup class) work as expected.

    For this the following will be checked:

    * fixture will be called correctly
    * parameters are given correctly (no automatic self for the first parameter - type will be checked too)
    * return value of fixture will be forwarded correctly

    The environment implements fixture in the definition scopes BALDERGLOB, SETUP and SCENARIO with all possible
    execution levels. It has one setup and one scenario class that use static method fixtures here. The test uses
    the following fixtures to secure that parameter and return values are managed correctly:

    **Class SetupA**:

    * ``fixture_setup_session`` (**non-static**) expects parameters ``balderglob_fixture_session``
    * ``fixture_setup_setup`` (**static**) expects parameters ``balderglob_fixture_setup`` and ``fixture_setup_session``
    * ``fixture_setup_scenario`` (**non-static**) expects parameters ``balderglob_fixture_scenario`` and
      ``fixture_setup_setup``
    * ``fixture_setup_variation`` (**static**) expects parameters ``balderglob_fixture_variation`` and
      ``fixture_scenario``
    * ``fixture_setup_testcase`` (**non-static**) expects parameters ``balderglob_fixture_testcase`` and
      ``fixture_setup_variation``

    **Class ScenarioA**:

    * ``fixture_scenario_session`` (**static**) expects parameters ``balderglob_fixture_session`` and
      ``fixture_setup_session``
    * ``fixture_scenario_setup`` (**non-static**) expects parameters ``balderglob_fixture_setup``,
      ``fixture_setup_setup`` and ``fixture_scenario_session``
    * ``fixture_scenario_scenario`` (**static**) expects parameters ``balderglob_fixture_scenario``,
      ``fixture_setup_scenario`` and ``fixture_scenario_setup``
    * ``fixture_scenario_variation`` (**non-static**) expects parameters ``balderglob_fixture_variation``,
      ``fixture_setup_variation`` and ``fixture_scenario_scenario``
    * ``fixture_scenario_testcase`` (**static**) expects parameters ``balderglob_fixture_testcase``,
      ``fixture_setup_testcase`` and ``fixture_scenario_variation``

    The testcase method validates the return values for all scenario and setup fixtures.
    """

    @property
    def expected_data(self) -> tuple:
        return (
            # FIXTURE-CONSTRUCTION: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_session
            {"cls": "SetupA", "meth": "fixture_setup_session", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_session
            {"cls": "ScenarioA", "meth": "fixture_scenario_session", "part": "construction"},
            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: balderglob_fixture_setup
            {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_setup
            {"cls": "SetupA", "meth": "fixture_setup_setup", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_setup
            {"cls": "ScenarioA", "meth": "fixture_scenario_setup", "part": "construction"},
            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_scenario
            {"cls": "SetupA", "meth": "fixture_setup_scenario", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_scenario
            {"cls": "ScenarioA", "meth": "fixture_scenario_scenario", "part": "construction"},
            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_variation
            {"cls": "SetupA", "meth": "fixture_setup_variation", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_variation
            {"cls": "ScenarioA", "meth": "fixture_scenario_variation", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_testcase
            {"cls": "SetupA", "meth": "fixture_setup_testcase", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_testcase
            {"cls": "ScenarioA", "meth": "fixture_scenario_testcase", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # TESTCASE: ScenarioA.test_a_1
            {"cls": "ScenarioA", "meth": "test_a_1"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: ScenarioA.fixture_testcase
            {"cls": "ScenarioA", "meth": "fixture_scenario_testcase", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: SetupA.fixture_testcase
            {"cls": "SetupA", "meth": "fixture_setup_testcase", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
            # FIXTURE-TEARDOWN: ScenarioA.fixture_variation
            {"cls": "ScenarioA", "meth": "fixture_scenario_variation", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: SetupA.fixture_variation
            {"cls": "SetupA", "meth": "fixture_setup_variation", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_variation
            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
            # FIXTURE-TEARDOWN: ScenarioA.fixture_scenario
            {"cls": "ScenarioA", "meth": "fixture_scenario_scenario", "part": "teardown"},
            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: SetupA.fixture_scenario
            {"cls": "SetupA", "meth": "fixture_setup_scenario", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_scenario
            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
            # FIXTURE-TEARDOWN: ScenarioA.fixture_setup
            {"cls": "ScenarioA", "meth": "fixture_scenario_setup", "part": "teardown"},
            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: SetupA.fixture_setup
            {"cls": "SetupA", "meth": "fixture_setup_setup", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_setup
            {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "teardown"},
            # FIXTURE-TEARDOWN: ScenarioA.fixture_session
            {"cls": "ScenarioA", "meth": "fixture_scenario_session", "part": "teardown"},
            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: SetupA.fixture_session
            {"cls": "SetupA", "meth": "fixture_setup_session", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "teardown"},
        )

    @staticmethod
    def validate_finished_session(session: BalderSession):
        # check result states everywhere (have to be SUCCESS everywhere
        assert session.executor_tree.executor_result == ResultState.SUCCESS, \
            "test session does not terminates with success"

        assert session.executor_tree.construct_result.result == ResultState.SUCCESS, \
            "global executor tree construct part does not set ResultState.SUCCESS"
        assert session.executor_tree.body_result.result == ResultState.SUCCESS, \
            "global executor tree body part does not set ResultState.SUCCESS"
        assert session.executor_tree.teardown_result.result == ResultState.SUCCESS, \
            "global executor tree teardown part does not set ResultState.SUCCESS"
        for cur_setup_executor in session.executor_tree.setup_executors:
            assert cur_setup_executor.executor_result == ResultState.SUCCESS, \
                "the setup executor does not have result SUCCESS"

            assert cur_setup_executor.construct_result.result == ResultState.SUCCESS
            assert cur_setup_executor.body_result.result == ResultState.SUCCESS
            assert cur_setup_executor.teardown_result.result == ResultState.SUCCESS

            for cur_scenario_executor in cur_setup_executor.scenario_executors:
                assert cur_scenario_executor.executor_result == ResultState.SUCCESS, \
                    "the scenario executor does not have result SUCCESS"

                assert cur_scenario_executor.construct_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.body_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.teardown_result.result == ResultState.SUCCESS

                for cur_variation_executor in cur_scenario_executor.variation_executors:
                    assert cur_variation_executor.executor_result == ResultState.SUCCESS, \
                        "the variation executor does not have result SUCCESS"

                    assert cur_variation_executor.construct_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.body_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.teardown_result.result == ResultState.SUCCESS

                    for cur_testcase_executor in cur_variation_executor.testcase_executors:
                        assert cur_testcase_executor.executor_result == ResultState.SUCCESS, \
                            "the testcase executor does not have result SUCCESS"

                        assert cur_testcase_executor.construct_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.body_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.teardown_result.result == ResultState.SUCCESS



from .feature_i import SetupFeatureI
from .feature_ii import SetupFeatureII
from .feature_of_relevance_lvl3 import SetupFeatureOfRelevanceLvl3

__all__ = [
    "SetupFeatureI",
    "SetupFeatureII",
    "SetupFeatureOfRelevanceLvl3"
]



from __future__ import annotations

from _balder.objects.connections.osi_1_physical import BluetoothConnection, CanBusConnection, CoaxialCableConnection, \
    DslConnection, RS232Connection, RS422Connection, RS485Connection, IsdnConnection, I2CConnection, I2SConnection, \
    OneWireConnection, OpticalFiberConnection, SpiConnection, TwistedPairCableConnection, UsbConnection, WifiConnection
from _balder.objects.connections.osi_2_datalink import EthernetConnection, WirelessLanConnection, LLDPConnection, \
    ProfibusConnection
from _balder.objects.connections.osi_3_network import IpSecConnection, IPv4Connection, IPv6Connection, \
    ICMPv4Connection, ICMPv6Connection, IPConnection, ICMPConnection
from _balder.objects.connections.osi_4_transport import TcpIPv4Connection, UdpIPv4Connection, TcpIPv6Connection, \
    UdpIPv6Connection, TcpConnection, UdpConnection
from _balder.objects.connections.osi_5_session import PptpConnection
from _balder.objects.connections.osi_6_presentation import TelnetConnection
from _balder.objects.connections.osi_7_application import HttpConnection, ImapConnection, LdapConnection, \
    NtpConnection, RpcConnection, SmtpConnection, SntpConnection, SshConnection, DnsConnection

__all__ = [
    # OSI: Physical Layer
    "BluetoothConnection", "CanBusConnection", "CoaxialCableConnection", "DslConnection", "RS232Connection",
    "RS422Connection", "RS485Connection", "IsdnConnection", "I2CConnection", "I2SConnection", "OneWireConnection",
    "OpticalFiberConnection", "SpiConnection", "TwistedPairCableConnection", "UsbConnection", "WifiConnection",

    # OSI: Data Link Layer
    "EthernetConnection", "WirelessLanConnection", "LLDPConnection", "ProfibusConnection",

    # OSI: Network Layer
    "IpSecConnection", "IPv4Connection", "IPv6Connection", "ICMPv4Connection", "ICMPv6Connection",
    "IPConnection", "ICMPConnection",

    # OSI: Transport Layer
    "TcpIPv4Connection", "UdpIPv4Connection", "TcpIPv6Connection", "UdpIPv6Connection", "TcpConnection",
    "UdpConnection",

    # OSI: Session Layer
    "PptpConnection",

    # OSI Presentation Layer
    "TelnetConnection",

    # OSI Application Layer
    "HttpConnection", "ImapConnection", "LdapConnection", "NtpConnection", "RpcConnection",
    "SmtpConnection", "SntpConnection", "SshConnection", "DnsConnection"
]



import balder
import logging

logger = logging.getLogger(__file__)


class MySettings(balder.BalderSettings):
    """This is our own setting object"""
    used_global_connection_tree = "some-never-used-custom-name"



import balder
import logging
from .setup_features import SetupFeatureI, SetupFeatureII
from ..lib.connections import BetweenConnection

logger = logging.getLogger(__name__)


class SetupA(balder.Setup):
    """This is a setup of category A (exactly the same as scenario A)"""

    class SetupDevice1(balder.Device):
        i = SetupFeatureI()

    @balder.connect(SetupDevice1, over_connection=BetweenConnection)
    class SetupDevice2(balder.Device):
        ii = SetupFeatureII()



from __future__ import annotations

import time
from typing import List, Dict, Union, Type, TYPE_CHECKING

import sys
import types
import traceback
from abc import ABC, abstractmethod
from _balder.testresult import FixturePartResult, ResultState
from _balder.previous_executor_mark import PreviousExecutorMark
from _balder.testresult import TestcaseResult

if TYPE_CHECKING:
    from _balder.setup import Setup
    from _balder.scenario import Scenario
    from _balder.fixture_manager import FixtureManager


class BasicExecutor(ABC):
    """
    The BasicExecutor class is an abstract class that represents the parent class of all executors. Together with other
    executor classes, an executor forms a tree structure in which individual tests, which later on are executed, are
    assigned to individual scenarios
    """
    # this property describes the runnable state of the executor branch before the executor is really used
    #  with this you can declare a whole branch as inactive, while the collecting process is active
    prev_mark = PreviousExecutorMark.RUNNABLE

    def __init__(self):
        # contains the result object for the CONSTRUCT FIXTURE part of this branch
        self.construct_result = FixturePartResult(self)
        # contains the result object for the BODY part of this branch (will be defined in subclasses)
        self.body_result = None
        # contains the result object for the TEARDOWN FIXTURE part of this branch
        self.teardown_result = FixturePartResult(self)

        # holds the execution time of this branch (with branch fixtures)
        self.execution_time_sec = None

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS ----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    @abstractmethod
    def all_child_executors(self) -> List[BasicExecutor]:
        """
        returns all child executors of this object or None if no child executors can exist (this element is a leaf)
        """

    @property
    @abstractmethod
    def parent_executor(self) -> BasicExecutor:
        """returns the parent executor of this element"""

    @property
    @abstractmethod
    def base_instance(self) -> object:
        """returns the base class instance to which this executor instance belongs or None if this element is a
        ExecutorTree"""

    @property
    @abstractmethod
    def fixture_manager(self) -> FixtureManager:
        """returns the active fixture manager instance"""

    @property
    def executor_result(self) -> ResultState:
        """
        This property returns the combined state of this executor object. This is determined by its properties
        ``construct_result``, ``body_result`` and ``teardown_result``.
        """
        relative_result = ResultState.NOT_RUN
        priority_order = ResultState.priority_order()

        if priority_order.index(self.construct_result.result) < priority_order.index(relative_result):
            relative_result = self.construct_result.result
        if priority_order.index(self.body_result.result) < priority_order.index(relative_result):
            relative_result = self.body_result.result
        if priority_order.index(self.teardown_result.result) < priority_order.index(relative_result):
            relative_result = self.teardown_result.result
        return relative_result

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    @abstractmethod
    def _prepare_execution(self):
        """
        This method runs before the branch will be executed and before the fixture construction code of this branch
        runs.
        """

    @abstractmethod
    def _body_execution(self):
        """
        This method runs between the fixture construction and teardown code. It should trigger the execution of the
        child branches.
        """

    @abstractmethod
    def _cleanup_execution(self):
        """
        This method runs after the branch was executed (also after the fixture teardown code ran)
        """

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def get_all_recognized_exception(self) -> List[Exception]:
        """
        This method returns all occurred Exception of this branch.

        :return: a list with all occurred and recorded exceptions
        """
        all_own = [self.construct_result.exception, self.body_result.exception, self.teardown_result.exception]
        if self.all_child_executors:
            for cur_child_executor in self.all_child_executors:
                all_own += cur_child_executor.get_all_recognized_exception()

        # filter all duplicated entries
        all_own = list(set(all_own))
        if None in all_own:
            all_own.remove(None)
        return all_own

    def set_result_for_whole_branch(self, value: ResultState):
        """
        This method sets the executor result for all sub executors.

        :param value: the new value that should be set for this branch
        """
        if value not in (ResultState.SKIP, ResultState.COVERED_BY, ResultState.NOT_RUN):
            raise ValueError("can not set a state that is not NOT_RUN, SKIP or COVERED_BY for a whole branch")
        for cur_child_executor in self.all_child_executors:
            if isinstance(cur_child_executor.body_result, TestcaseResult):
                cur_child_executor.body_result.set_result(result=value, exception=None)

    def has_runnable_elements(self) -> bool:
        """
        This method returns true if this executor element is runnable. The method returns true if this element has
        `prev_mark=RUNNABLE` and minimum one of its children has `prev_mark=RUNNABLE` too.
        """
        if self.prev_mark != PreviousExecutorMark.RUNNABLE:
            return False
        for cur_child in self.all_child_executors:
            if cur_child.prev_mark == PreviousExecutorMark.RUNNABLE:
                return True
        return False

    def get_all_base_instances_of_this_branch(
            self, with_type: Union[Type[Setup], Type[Scenario], Type[types.FunctionType]],
            only_runnable_elements: bool = True) -> List[Union[Setup, Scenario, object]]:
        """
        This method returns a list of all base instance elements that are from given type `with_type`. If `with_type`
        is a base type of this branch element, the method always returns a list with one element. If the base instance
        element is a child of the current branch, it could be a list with zero to infinite elements.

        .. note::
            The method only returns unique objects.
        """
        # search all higher classes and this one (if there is a match, return value is a list with exactly one element)
        cur_executor = self
        # only go through cur_executor == ExecutorTree (do not iterate with this object)
        while cur_executor.parent_executor is not None:
            if isinstance(cur_executor.base_instance, with_type):
                if not only_runnable_elements or cur_executor.has_runnable_elements():
                    return [cur_executor.base_instance]
            cur_executor = cur_executor.parent_executor

        # go in the branch and search in all children (recursively)
        result = []
        # if there are no child executors -> we do not need to search in children because there are no children
        if self.all_child_executors is not None:
            for cur_child_executor in self.all_child_executors:
                child_result = cur_child_executor.get_all_base_instances_of_this_branch(
                    with_type=with_type, only_runnable_elements=only_runnable_elements)
                result += child_result
        # remove duplicate items
        return list(set(result))

    @abstractmethod
    def cleanup_empty_executor_branches(self):
        """
        This method searches the whole tree and removes branches where an executor item has no own children. It can
        remove these branches, because they have no valid matchings.
        """

    def filter_tree_for_user_filters(self):
        """
        This method calls all user defined filters that are to be applied to the executor tree.
        """
        for cur_child_executor in self.all_child_executors:
            cur_child_executor.filter_tree_for_user_filters()

    def testsummary(self) -> Dict[ResultState, int]:
        """
        returns a dictionary with all possible :class:`ResultState` as keys and the number of times they have occurred
        in this branch as values
        """

        summary = {
            ResultState.NOT_RUN: 0,
            ResultState.FAILURE: 0,
            ResultState.ERROR: 0,
            ResultState.SUCCESS: 0,
            ResultState.SKIP: 0,
            ResultState.COVERED_BY: 0
        }

        if isinstance(self.body_result, TestcaseResult):
            summary[self.executor_result] = 1
        else:
            for cur_child_exec in self.all_child_executors:
                cur_child_dict = cur_child_exec.testsummary()
                for cur_key in cur_child_dict.keys():
                    if cur_key in summary:
                        summary[cur_key] += cur_child_dict[cur_key]
                    else:
                        summary[cur_key] = cur_child_dict[cur_key]
        return summary

    def execute(self):
        """
        Executes the whole branch
        """
        start_time = time.perf_counter()
        self._prepare_execution()

        try:
            try:
                self.fixture_manager.enter(self)
                self.construct_result.set_result(ResultState.SUCCESS)

                self._body_execution()
            except Exception as exc:
                # this has to be a construction fixture error
                traceback.print_exception(*sys.exc_info())
                self.construct_result.set_result(ResultState.ERROR, exc)
            finally:
                if self.fixture_manager.is_allowed_to_leave(self):
                    self.fixture_manager.leave(self)
                    self.teardown_result.set_result(ResultState.SUCCESS)
        except Exception as exc:
            # this has to be a teardown fixture error
            traceback.print_exception(*sys.exc_info())
            self.teardown_result.set_result(ResultState.ERROR, exc)

        self._cleanup_execution()

        self.execution_time_sec = time.perf_counter() - start_time



import logging
import balder
from ..lib.connections import MySimplySharedMemoryConnection
from ..lib.utils import SharedObj
from .features_setup import PyAddCalculate, PyAddProvideANumber
from balder.exceptions import AccessToUnmappedVDeviceException

logger = logging.getLogger(__name__)


class SetupPythonAddVDevice2WithTwoMapped(balder.Setup):

    class Calculator(balder.Device):
        calc = PyAddCalculate(ANumberProviderDevice2="NumberProviderTwo")

    @balder.connect(Calculator, over_connection=MySimplySharedMemoryConnection)
    class NumberProviderOne(balder.Device):
        n = PyAddProvideANumber()

    @balder.connect(Calculator, over_connection=MySimplySharedMemoryConnection)
    class NumberProviderTwo(balder.Device):
        n = PyAddProvideANumber()

    @balder.fixture(level="testcase")
    def cleanup_memory(self):
        SharedObj.shared_mem_list = []

    def validate_vdevice_references(self):
        assert isinstance(self.Calculator.calc.get_number_of_provider_2(), int), \
            f"the currently mapped internal vdevice returns an unexpected value " \
            f"`{self.Calculator.calc.get_number_of_provider_2()}`"

        try:
            self.Calculator.calc.get_number_of_provider_1()
            assert False, "provider 1 does return a value -> not allowed, because this VDevice was not mapped"
        except AccessToUnmappedVDeviceException as exc:
            logger.info(f'access to unmapped VDevice throws error message: `{str(exc)}`')
            assert exc.args[0] == "it is not allowed to access the attributes of an unmapped VDevice - " \
                                  "did you forget to map it?"

    @balder.fixture(level='variation')
    def fixture_vdevice_access_variation(self):
        self.validate_vdevice_references()
        yield
        self.validate_vdevice_references()

    @balder.fixture(level='testcase')
    def fixture_vdevice_access_testcase(self):
        self.validate_vdevice_references()
        yield
        self.validate_vdevice_references()



import balder
import logging
from ..lib.features import FeatureI, FeatureII
from ..lib.connections import AConnection
from ..lib.utils import FixtureReturn
from ..balderglob import RuntimeObserver

logger = logging.getLogger(__file__)


class ScenarioA(balder.Scenario):
    """This is the scenario of category A"""

    class ScenarioDevice1(balder.Device):
        i = FeatureI()

    @balder.connect(ScenarioDevice1, over_connection=AConnection)
    class ScenarioDevice2(balder.Device):
        ii = FeatureII()

    def test_a_1(self, fixture_scenario_session, fixture_scenario_setup, fixture_scenario_scenario,
                 fixture_scenario_variation, fixture_scenario_testcase, fixture_setup_session, fixture_setup_setup,
                 fixture_setup_scenario, fixture_setup_variation, fixture_setup_testcase):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.test_a_1, category="testcase",
                                  msg=f"execute Test `{ScenarioA.test_a_1.__qualname__}`")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        assert isinstance(self, ScenarioA)
        assert fixture_scenario_session == FixtureReturn.SCENARIO_SESSION
        assert fixture_scenario_setup == FixtureReturn.SCENARIO_SETUP
        assert fixture_scenario_scenario == FixtureReturn.SCENARIO_SCENARIO
        assert fixture_scenario_variation == FixtureReturn.SCENARIO_VARIATION
        assert fixture_scenario_testcase == FixtureReturn.SCENARIO_TESTCASE
        assert fixture_setup_session == FixtureReturn.SETUP_SESSION
        assert fixture_setup_setup == FixtureReturn.SETUP_SETUP
        assert fixture_setup_scenario == FixtureReturn.SETUP_SCENARIO
        assert fixture_setup_variation == FixtureReturn.SETUP_VARIATION
        assert fixture_setup_testcase == FixtureReturn.SETUP_TESTCASE

    @balder.fixture(level="session")
    def fixture_scenario_session(self, balderglob_fixture_session):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_scenario_session, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        assert isinstance(self, ScenarioA)
        assert balderglob_fixture_session == FixtureReturn.BALDERGLOB_SESSION

        yield FixtureReturn.SCENARIO_SESSION

        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_scenario_session, category="fixture",
                                  part="teardown", msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        assert isinstance(self, ScenarioA)
        assert balderglob_fixture_session == FixtureReturn.BALDERGLOB_SESSION

    @classmethod
    @balder.fixture(level="setup")
    def fixture_scenario_setup(cls, balderglob_fixture_setup, fixture_setup_setup, fixture_scenario_session):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_scenario_setup, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        cls.ScenarioDevice1.i.do_something()
        cls.ScenarioDevice2.ii.do_something()

        assert issubclass(cls, ScenarioA)
        assert balderglob_fixture_setup == FixtureReturn.BALDERGLOB_SETUP
        assert fixture_scenario_session == FixtureReturn.SCENARIO_SESSION
        assert fixture_setup_setup == FixtureReturn.SETUP_SETUP

        yield FixtureReturn.SCENARIO_SETUP

        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_scenario_setup, category="fixture",
                                  part="teardown", msg="begin execution TEARDOWN of fixture")

        cls.ScenarioDevice1.i.do_something()
        cls.ScenarioDevice2.ii.do_something()

        assert issubclass(cls, ScenarioA)
        assert balderglob_fixture_setup == FixtureReturn.BALDERGLOB_SETUP
        assert fixture_scenario_session == FixtureReturn.SCENARIO_SESSION
        assert fixture_setup_setup == FixtureReturn.SETUP_SETUP

    @balder.fixture(level="scenario")
    def fixture_scenario_scenario(self, balderglob_fixture_scenario, fixture_setup_scenario, fixture_scenario_setup):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_scenario_scenario, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        assert isinstance(self, ScenarioA)
        assert balderglob_fixture_scenario == FixtureReturn.BALDERGLOB_SCENARIO
        assert fixture_scenario_setup == FixtureReturn.SCENARIO_SETUP
        assert fixture_setup_scenario == FixtureReturn.SETUP_SCENARIO

        yield FixtureReturn.SCENARIO_SCENARIO

        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_scenario_scenario, category="fixture",
                                  part="teardown", msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        assert isinstance(self, ScenarioA)
        assert balderglob_fixture_scenario == FixtureReturn.BALDERGLOB_SCENARIO
        assert fixture_scenario_setup == FixtureReturn.SCENARIO_SETUP
        assert fixture_setup_scenario == FixtureReturn.SETUP_SCENARIO

    @classmethod
    @balder.fixture(level="variation")
    def fixture_scenario_variation(cls, balderglob_fixture_variation, fixture_setup_variation,
                                   fixture_scenario_scenario):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_scenario_variation, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        cls.ScenarioDevice1.i.do_something()
        cls.ScenarioDevice2.ii.do_something()

        assert issubclass(cls, ScenarioA)
        assert balderglob_fixture_variation == FixtureReturn.BALDERGLOB_VARIATION
        assert fixture_scenario_scenario == FixtureReturn.SCENARIO_SCENARIO
        assert fixture_setup_variation == FixtureReturn.SETUP_VARIATION

        yield FixtureReturn.SCENARIO_VARIATION

        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_scenario_variation, category="fixture",
                                  part="teardown", msg="begin execution TEARDOWN of fixture")

        cls.ScenarioDevice1.i.do_something()
        cls.ScenarioDevice2.ii.do_something()

        assert issubclass(cls, ScenarioA)
        assert balderglob_fixture_variation == FixtureReturn.BALDERGLOB_VARIATION
        assert fixture_scenario_scenario == FixtureReturn.SCENARIO_SCENARIO
        assert fixture_setup_variation == FixtureReturn.SETUP_VARIATION

    @balder.fixture(level="testcase")
    def fixture_scenario_testcase(self, balderglob_fixture_testcase, fixture_setup_testcase, fixture_scenario_variation):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_scenario_testcase, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        assert isinstance(self, ScenarioA)
        assert balderglob_fixture_testcase == FixtureReturn.BALDERGLOB_TESTCASE
        assert fixture_scenario_variation == FixtureReturn.SCENARIO_VARIATION
        assert fixture_setup_testcase == FixtureReturn.SETUP_TESTCASE

        yield FixtureReturn.SCENARIO_TESTCASE

        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_scenario_testcase, category="fixture",
                                  part="teardown", msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        assert isinstance(self, ScenarioA)
        assert balderglob_fixture_testcase == FixtureReturn.BALDERGLOB_TESTCASE
        assert fixture_scenario_variation == FixtureReturn.SCENARIO_VARIATION
        assert fixture_setup_testcase == FixtureReturn.SETUP_TESTCASE



import balder
from ..balderglob import RuntimeObserver


class FeatureI(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureI, FeatureI.do_something, "enter `FeatureI.do_something`", category="feature")


class FeatureII(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureII, FeatureII.do_something, "enter `FeatureII.do_something`", category="feature")


class FeatureIChild(FeatureI):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureIChild, FeatureIChild.do_something,
            "enter `FeatureIChild.do_something`", category="feature")


class FeatureNewI(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureNewI, FeatureNewI.do_something,
            "enter `FeatureNewI.do_something`", category="feature")


class FeatureOfRelevanceLvl1(balder.Feature):

    class VDeviceWithI(balder.VDevice):
        i = FeatureI()

    class VDeviceWithII(balder.VDevice):
        i = FeatureII()

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureOfRelevanceLvl1, FeatureOfRelevanceLvl1.do_something,
            "enter `FeatureOfRelevanceLvl1.do_something`", category="feature")


class FeatureOfRelevanceLvl2(FeatureOfRelevanceLvl1):
    """this feature will be used in scenario"""

    class VDeviceWithI(FeatureOfRelevanceLvl1.VDeviceWithI):
        # overwrites the old one
        i = FeatureIChild()
        # defines a new one
        new = FeatureNewI()

    class VDeviceWithII(FeatureOfRelevanceLvl1.VDeviceWithII):
        i = FeatureII()

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureOfRelevanceLvl2, FeatureOfRelevanceLvl2.do_something,
            "enter `FeatureOfRelevanceLvl2.do_something`", category="feature")



from __future__ import annotations
from typing import TYPE_CHECKING, Type, Tuple, List, Union

import pathlib

if TYPE_CHECKING:
    from _balder.executor.executor_tree import ExecutorTree
    from _balder.balder_session import BalderSession
    from _balder.setup import Setup
    from _balder.scenario import Scenario
    import argparse


class BalderPlugin:
    """
    This is the balder plugin class. You can create your own plugin, by creating a subclass of it. With that you are
    able to overwrite the methods you want to use in your plugin.
    """

    def __init__(self, session: BalderSession):
        self.balder_session = session

    def addoption(self, argument_parser: argparse.ArgumentParser):
        """
        The callback will be executed while the `ArgumentParser` is being created.

        :param argument_parser: the argument parser object
        """

    def modify_collected_pyfiles(self, pyfiles: List[pathlib.Path]) -> List[pathlib.Path]:
        """
        This callback will be executed after the :class:`Collector` has collected all python files that are inside the
        current working directory.

        .. note::
            Note that these files are not filtered yet. The list will contain every existing python file.

        :param pyfiles: a list with all python filepaths

        :return: the new list of all filepaths
        """

    def modify_collected_classes(self, scenarios: List[Type[Scenario]], setups: List[Type[Setup]]) \
            -> Tuple[List[Type[Scenario]], List[Type[Setup]]]:
        """
        This callback will be executed after the :class:`Collector` has collected the :class:`Scenario` classes and the
        :class:`Setup` classes.

        :param scenarios: all collected :class:`Scenario` classes that are currently in the collected list

        :param setups: all collected :class:`Setup` classes that are currently in the collected list

        :return: a tuple of lists, where the first list is the new list with all :class:`Scenario` classes, the second
                 element is a list with all :class:`Setup` classes
        """

    def filter_executor_tree(self, executor_tree: ExecutorTree) -> None:
        """
        This callback will be executed before the ExecutorTree runs. It contains the current representation
        of the :class:`ExecutorTree`, that will be executed in the next step. With this callback it is possible to
        manipulate the :class:`ExecutorTree`. You have not to return something, the given ``executor_tree`` is a
        reference.

        :param executor_tree: the reference to the main :class:`ExecutorTree` object balder uses for this session
        """

    def session_finished(self, executor_tree: Union[ExecutorTree, None]):
        """
        This callback will be executed at the end of every session. The callback will run in a `collect-only` and
        `resolve-only` session too. Note, that the `executor_tree` argument is None in a `collect-only` session.

        :param executor_tree: the reference to the main :class:`ExecutorTree` object that balder uses for this session
        """



from __future__ import annotations
from typing import Dict, Type, Union, TYPE_CHECKING

import logging
import inspect
from abc import ABC, abstractmethod
from _balder.controllers.controller import Controller
from _balder.feature import Feature

if TYPE_CHECKING:
    from _balder.scenario import Scenario
    from _balder.setup import Setup

logger = logging.getLogger(__file__)


class BaseDeviceController(Controller, ABC):
    """
    This is the abstract controller class for :class:`Device` items.
    """
    def __init__(self):

        #: contains the original instantiated objects for the related device class (will be automatically set by
        #:  :class:`Collector`)
        self._original_instanced_features: Union[Dict[str, Feature], None] = {}

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS -----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    @abstractmethod
    def get_outer_class(self) -> Union[Type[Scenario], Type[Setup], None]:
        """
        This method delivers the outer class of this device. This has to be a :meth:`Setup` or a :meth:`Scenario`.
        """

    def get_all_instantiated_feature_objects(self) -> Dict[str, Feature]:
        """
        This method returns all instantiated :meth:`Feature` classes that were defined as static attributes within the
        related device.

        :return: supplies a dictionary with the name of the attribute as key and the current feature class as value
        """

        results = {}
        for cur_attr_name, cur_elem in inspect.getmembers(self.related_cls, lambda elem: isinstance(elem, Feature)):
            results[cur_attr_name] = cur_elem
        return results

    def get_original_instanced_feature_objects(self) -> Dict[str, Feature]:
        """
        This method returns the original instanced feature objects of the related device
        """
        if self._original_instanced_features is None:
            # todo we should use a balder exception here!!
            raise EnvironmentError('can not access the original instantiated features before they were set with '
                                   '`save_all_original_instanced_features`')
        return self._original_instanced_features

    def save_all_original_instanced_features(self):
        """
        This property sets the internal dictionary about the original instantiated features of this
        :class:`Device`/:class:`VDevice`. This is done, to ensure that balder has saved an original copy of the original
        instantiated abstract features. The real features will be overwritten for each new variation by the
        :class:`ExecutorTree`!
        """
        self._original_instanced_features = self.get_all_instantiated_feature_objects()



from __future__ import annotations
from typing import Union, List, Tuple, Dict, Type, TYPE_CHECKING

import os
import sys
import inspect
import pathlib
import argparse
import balder
from _balder.balder_plugin import BalderPlugin
from _balder.plugin_manager import PluginManager
from _balder.executor.executor_tree import ExecutorTree
from _balder.collector import Collector
from _balder.solver import Solver
from _balder.exceptions import DuplicateBalderSettingError
from _balder.balder_settings import BalderSettings

if TYPE_CHECKING:
    from _balder.setup import Setup
    from _balder.device import Device
    from _balder.scenario import Scenario
    from _balder.connection import Connection


class BalderSession:
    """
    This is the main balder executable object. It contains all information about the current session and executes the
    different steps.

    This object contains all command line arguments that can be used, while calling `balder ..`. All settings that
    are given in `balderglob.py` will be imported in `self.baldersetting`
    """
    # this is the default value (will be overwritten in this constructor if necessary)
    baldersettings: BalderSettings = BalderSettings()

    def __init__(self, cmd_args: Union[List[str], None] = None, working_dir: Union[pathlib.Path, None] = None):
        """

        :param cmd_args: optional the command line list of strings that should be parsed instead of parsing the real
                         command line arguments (used for testing)

        :param working_dir: the working directory that should be used instead of the given value in `cmd_arg_str` or
                            the current directory (determined by `os.getcwd()`)
        """
        #: contains the alternative command line arguments as a string list (has to be given, if the object should
        #: not use the console params of this call)
        self._alt_cmd_args = cmd_args
        #: contains a reference to the ArgumentParser that parses the command line string
        self.cmd_arg_parser = None
        #: contains the parsed object
        self.parsed_args: argparse.Namespace

        ##
        # all general settings that can be modified by command line arguments
        ##

        #: the working directory for this balder session (default: current directory from `os.getcwd()`)
        self.working_dir: Union[pathlib.Path, None] = pathlib.Path(os.getcwd())
        #: specifies that the tests should only be collected but not be resolved and executed
        self.collect_only: Union[bool, None] = None
        #: specifies that the tests should only be collected and resolved but not executed
        self.resolve_only: Union[bool, None] = None
        #: contains a number of :class:`Setup` class strings that should only be considered for the execution
        self.only_with_setup: Union[List[str], None] = None
        #: contains a number of :class:`Scenario` class strings that should only be considered for the execution
        self.only_with_scenario: Union[List[str], None] = None
        #: if this is true, the test run should include duplicated tests that are declared as covered_by another test
        #: method
        self.force_covered_by_duplicates: Union[bool, None] = None

        self.preparse_args()

        #: overwrite working directory if necessary
        if working_dir:
            self.working_dir = working_dir

        ##
        # instantiate and initialize all components to completely load the plugins (plugins could access the command
        # line argument parser)
        ##
        #: contains the reference to the used PluginManager
        self.plugin_manager = PluginManager()
        #: contains the reference to the used :class:`Collector` class
        self.collector = Collector(self.working_dir)

        # determine the balder settings
        BalderSession.baldersettings = self.get_baldersettings_from_balderglob()
        BalderSession.baldersettings = BalderSession.baldersettings if BalderSession.baldersettings is not None else \
            BalderSettings()

        if BalderSession.baldersettings.force_covered_by_duplicates:
            # overwrite console argument only if the value in BalderSettings is true (because cmd line can only
            # overwrite the value False)
            self.force_covered_by_duplicates = True

        for cur_plugin_cls in self.get_balderplugins_from_balderglob():
            self.plugin_manager.register(cur_plugin_cls, self)

        ##
        # instantiate all sub objects that are relevant for this test session
        ##
        self.parse_args()

        #: overwrite working directory if necessary
        if working_dir:
            self.working_dir = working_dir

        #: contains the reference to the used :class:`Solver´ class (or none, if there was no solving executed till
        #: now)
        self.solver: Union[Solver, None] = None
        #: contains the reference to the used :class:`ExecutorTree` class (or none, if there was no solving executed
        #: till now)
        self.executor_tree: Union[ExecutorTree, None] = None

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    @staticmethod
    def get_current_active_global_conntree_name():
        """
        This method returns the current active global connection tree name, depending on the current active setting in
        :class:`BalderSettings`, that is active for the current run.
        """
        if BalderSession.baldersettings is None:
            raise ValueError("no baldersettings loaded yet")

        return BalderSession.baldersettings.used_global_connection_tree

    # ---------------------------------- CLASS METHODS -----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def all_collected_pyfiles(self) -> List[pathlib.Path]:
        """returns all collected pyfiles"""
        try:
            return self.collector.all_pyfiles
        except AttributeError as exc:
            raise RuntimeError("this property is only available after the collecting process was executed") from exc

    @property
    def all_collected_setups(self) -> List[Type[Setup]]:
        """returns all collected :class:`Setup` classes"""
        try:
            return self.collector.all_setups
        except AttributeError as exc:
            raise RuntimeError("this property is only available after the collecting process was executed") from exc

    @property
    def all_collected_scenarios(self) -> List[Type[Scenario]]:
        """returns all collected :class:`Scenario` classes"""
        try:
            return self.collector.all_scenarios
        except AttributeError as exc:
            raise RuntimeError("this property is only available after the collecting process was executed") from exc

    @property
    def all_collected_connections(self) -> List[Type[Connection]]:
        """returns all collected :class:`Connection` classes"""
        try:
            return self.collector.all_connections
        except AttributeError as exc:
            raise RuntimeError("this property is only available after the collecting process was executed") from exc

    @property
    def all_resolved_mappings(self) -> List[Tuple[Type[Setup], Type[Scenario], Dict[Type[Device], Type[Device]]]]:
        """returns all resolved mappings for the :class:`Device` mappings between :class:`Scenario` and
        :class:`Setup`"""
        try:
            return self.solver.all_mappings
        except AttributeError as exc:
            raise RuntimeError("this property is only available after the resolving process was executed") from exc

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def get_baldersettings_from_balderglob(self) -> Union[BalderSettings, None]:
        """
        Helper method that checks if there is a valid :class:`BalderSettings` class in the given module
        """
        module = self.collector.load_balderglob_py_file()
        class_members = inspect.getmembers(module, inspect.isclass)
        all_classes = []
        for _, cur_class in class_members:
            if issubclass(cur_class, BalderSettings):
                all_classes.append(cur_class)
        if len(all_classes) == 0:
            return None
        if len(all_classes) > 1:
            raise DuplicateBalderSettingError(f"found more than one object that could be a BalderSettings object - "
                                              f"found {','.join([cur_class.__name__ for cur_class in all_classes])}")
        return all_classes[0]()

    def get_balderplugins_from_balderglob(self) -> List[Type[BalderPlugin]]:
        """
        Helper method that loads all valid :class:`BalderPlugin` classes and returns them
        """
        module = self.collector.load_balderglob_py_file()
        class_members = inspect.getmembers(module, inspect.isclass)
        all_classes = []
        for _, cur_class in class_members:
            if issubclass(cur_class, BalderPlugin):
                all_classes.append(cur_class)
        return all_classes

    def preparse_args(self):
        """
        This method pre-parses the console arguments and checks if one or more of the most important (for example
        `--working-dir`) attributes are given as CLI argument. This will be automatically pre-set. With that balder
        secures that it can read the correct `balderglob.py` file, to initialize the plugins correctly.
        """
        argv_working_dir_key = "--working-dir"
        if argv_working_dir_key in sys.argv:
            found_idx = sys.argv.index(argv_working_dir_key)
            if found_idx != -1:
                if len(sys.argv) <= found_idx + 1:
                    raise AttributeError(f"no path given for `{argv_working_dir_key}`")

                self.working_dir = pathlib.Path(sys.argv[found_idx + 1]).absolute()
                if not self.working_dir.is_dir():
                    raise NotADirectoryError(
                        f'can not parse the given working directory `{self.working_dir}` correctly or the given '
                        f'path is no directory..')

    def parse_args(self):
        """
        This method can be used to parse the `arg` object and fill all data from that object into the properties of this
        :class:`BalderSession` object.
        """
        self.cmd_arg_parser = argparse.ArgumentParser(
            description='Balder is a simple scenario-based test system that allows you to run your tests on various '
                        'devices without rewriting them')

        self.cmd_arg_parser.add_argument(
            '--working-dir', nargs="?", default=os.getcwd(),
            help="a explicit working directory on which the testsystem is to be executed with")

        self.cmd_arg_parser.add_argument(
            '--collect-only', action='store_true',
            help="specifies that the tests are only collected but not resolved and executed")

        self.cmd_arg_parser.add_argument(
            '--resolve-only', action='store_true',
            help="specifies that the tests are only collected and resolved but not executed")

        self.cmd_arg_parser.add_argument(
            '--only-with-setup', nargs="*",
            help="defines a number of Setup classes which should only be considered for the execution")

        self.cmd_arg_parser.add_argument(
            '--only-with-scenario', nargs="*",
            help="defines a number of Scenario classes which should only be considered for the execution")
        self.cmd_arg_parser.add_argument(
            '--force-covered-by-duplicates', action='store_true',
            help="specifies that the test run should include duplicated tests that are declared as covered_by another "
                 "test method (also true if it was already set in baldersetting object)")

        self.plugin_manager.execute_addoption(self.cmd_arg_parser)

        self.parsed_args = self.cmd_arg_parser.parse_args(self._alt_cmd_args)

        if self.parsed_args.only_with_scenario is not None:
            raise NotImplementedError("the parameter `--only-with-scenario` is not supported in this version of balder")
        if self.parsed_args.only_with_setup is not None:
            raise NotImplementedError("the parameter `--only-with-setup` is not supported in this version of balder")

        self.working_dir = self.parsed_args.working_dir
        self.collect_only = self.parsed_args.collect_only
        self.resolve_only = self.parsed_args.resolve_only
        self.only_with_setup = self.parsed_args.only_with_setup
        self.only_with_scenario = self.parsed_args.only_with_scenario
        self.force_covered_by_duplicates = self.parsed_args.force_covered_by_duplicates

    def collect(self):
        """
        This method collects all data.
        """
        self.collector.collect(plugin_manager=self.plugin_manager)

    def solve(self):
        """
        This method resolves all classes and executes different checks, that can be done before the test session starts.
        """
        self.solver = Solver(setups=self.all_collected_setups, scenarios=self.all_collected_scenarios,
                             connections=self.all_collected_connections, raw_fixtures=self.collector.raw_fixtures)
        self.solver.resolve(plugin_manager=self.plugin_manager)

    def create_executor_tree(self):
        """
        This method creates the executor tree object.

        .. note::
            Note that the method creates an :class:`ExecutorTree`, that hasn't to be completely resolved yet.
        """
        self.executor_tree = self.solver.get_executor_tree(plugin_manager=self.plugin_manager)
        self.plugin_manager.execute_filter_executor_tree(executor_tree=self.executor_tree)

    def execute_executor_tree(self):
        """
        This method executes the :class:`ExecutorTree`.
        """
        self.executor_tree.execute()

    def run(self):
        """
        This method executes the whole session
        """
        line_length = 120

        self.collect()

        def print_rect_row(text):
            line = "| " + text
            line = line + " " * (line_length - len(line) - 1) + "|"
            print(line)

        print("+" + "-" * (line_length - 2) + "+")
        print_rect_row("BALDER Testsystem")
        sys_version = sys.version.replace('\n', '')
        print_rect_row(f" python version {sys_version} | balder version {balder.__version__}")
        print("+" + "-" * (line_length - 2) + "+")
        print(f"Collect {len(self.all_collected_setups)} Setups and {len(self.all_collected_scenarios)} Scenarios")
        if not self.collect_only:
            self.solve()
            self.create_executor_tree()
            print(f"  resolve them to {len(self.executor_tree.get_all_variation_executors())} mapping candidates")
            print("")
            if not self.resolve_only:
                self.execute_executor_tree()
            else:
                self.executor_tree.print_tree()

        self.plugin_manager.execute_session_finished(self.executor_tree)



from _balder.testresult import ResultState
from _balder.balder_session import BalderSession

from tests.test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0ExactlyOneMoreFeatureInSetup(Base0EnvtesterClass):
    """
    This testcase uses a reduced version of the basic envtester environment. It should have one matching between
    the ``ScenarioA`` and the ``SetupA`` because all scenario-device-features are implemented. The test secures that
    it will be assigned specially because the ``SetupDevice1`` has one additional feature ``FeatureV``, that is not
    implemented in any scenario device.
    """

    @property
    def expected_data(self) -> tuple:
        return (
            # FIXTURE-CONSTRUCTION: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_session
            {"cls": "SetupA", "meth": "fixture_session", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureV", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_session
            {"cls": "ScenarioA", "meth": "fixture_session", "part": "construction"},
            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: balderglob_fixture_setup
            {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_setup
            {"cls": "SetupA", "meth": "fixture_setup", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureV", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_setup
            {"cls": "ScenarioA", "meth": "fixture_setup", "part": "construction"},
            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_scenario
            {"cls": "SetupA", "meth": "fixture_scenario", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureV", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_scenario
            {"cls": "ScenarioA", "meth": "fixture_scenario", "part": "construction"},
            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_variation
            {"cls": "SetupA", "meth": "fixture_variation", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureV", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_variation
            {"cls": "ScenarioA", "meth": "fixture_variation", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            [
                (
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_testcase
                    {"cls": "SetupA", "meth": "fixture_testcase", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioA.fixture_testcase
                    {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # TESTCASE: ScenarioA.test_a_1
                    {"cls": "ScenarioA", "meth": "test_a_1"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: ScenarioA.fixture_testcase
                    {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupA.fixture_testcase
                    {"cls": "SetupA", "meth": "fixture_testcase", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_testcase
                    {"cls": "SetupA", "meth": "fixture_testcase", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioA.fixture_testcase
                    {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # TESTCASE: ScenarioA.test_a_2
                    {"cls": "ScenarioA", "meth": "test_a_2"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: ScenarioA.fixture_testcase
                    {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupA.fixture_testcase
                    {"cls": "SetupA", "meth": "fixture_testcase", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                ),
            ],
            # FIXTURE-TEARDOWN: ScenarioA.fixture_variation
            {"cls": "ScenarioA", "meth": "fixture_variation", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: SetupA.fixture_variation
            {"cls": "SetupA", "meth": "fixture_variation", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureV", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_variation
            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
            # FIXTURE-TEARDOWN: ScenarioA.fixture_scenario
            {"cls": "ScenarioA", "meth": "fixture_scenario", "part": "teardown"},
            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: SetupA.fixture_scenario
            {"cls": "SetupA", "meth": "fixture_scenario", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureV", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_scenario
            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
            # FIXTURE-TEARDOWN: ScenarioA.fixture_setup
            {"cls": "ScenarioA", "meth": "fixture_setup", "part": "teardown"},
            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: SetupA.fixture_setup
            {"cls": "SetupA", "meth": "fixture_setup", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureV", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_setup
            {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "teardown"},
            # FIXTURE-TEARDOWN: ScenarioA.fixture_session
            {"cls": "ScenarioA", "meth": "fixture_session", "part": "teardown"},
            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: SetupA.fixture_session
            {"cls": "SetupA", "meth": "fixture_session", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureV", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "teardown"},
        )

    @staticmethod
    def validate_finished_session(session: BalderSession):
        # check result states everywhere (have to be SUCCESS everywhere)
        assert session.executor_tree.executor_result == ResultState.SUCCESS, \
            "test session does not terminates with success"

        assert session.executor_tree.construct_result.result == ResultState.SUCCESS, \
            "global executor tree construct part does not set ResultState.SUCCESS"
        assert session.executor_tree.body_result.result == ResultState.SUCCESS, \
            "global executor tree body part does not set ResultState.SUCCESS"
        assert session.executor_tree.teardown_result.result == ResultState.SUCCESS, \
            "global executor tree teardown part does not set ResultState.SUCCESS"
        for cur_setup_executor in session.executor_tree.setup_executors:
            assert cur_setup_executor.executor_result == ResultState.SUCCESS, \
                "the setup executor does not have result SUCCESS"

            assert cur_setup_executor.construct_result.result == ResultState.SUCCESS
            assert cur_setup_executor.body_result.result == ResultState.SUCCESS
            assert cur_setup_executor.teardown_result.result == ResultState.SUCCESS

            for cur_scenario_executor in cur_setup_executor.scenario_executors:
                assert cur_scenario_executor.executor_result == ResultState.SUCCESS, \
                    "the scenario executor does not have result SUCCESS"

                assert cur_scenario_executor.construct_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.body_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.teardown_result.result == ResultState.SUCCESS

                for cur_variation_executor in cur_scenario_executor.variation_executors:
                    assert cur_variation_executor.executor_result == ResultState.SUCCESS, \
                        "the variation executor does not have result SUCCESS"

                    assert cur_variation_executor.construct_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.body_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.teardown_result.result == ResultState.SUCCESS

                    for cur_testcase_executor in cur_variation_executor.testcase_executors:
                        assert cur_testcase_executor.executor_result == ResultState.SUCCESS, \
                            "the testcase executor does not have result SUCCESS"

                        assert cur_testcase_executor.construct_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.body_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.teardown_result.result == ResultState.SUCCESS



from __future__ import annotations
from typing import Literal

import functools
from _balder.collector import Collector
from _balder.fixture_manager import FixtureManager


def fixture(level: Literal['session', 'setup', 'scenario', 'variation', 'testcase']):
    """
    This decorator declares the decorated function/method as a fixture function/method.

    :param level: the execution level the fixture should have
    """
    allowed_levels = FixtureManager.EXECUTION_LEVEL_ORDER

    if level not in allowed_levels:
        raise ValueError(f"the value of `level` must be a `str` with one of the values `{'`, `'.join(allowed_levels)}`")

    def decorator_fixture(func):
        # always add the fixture to FixtureManager.raw_fixtures - class determination will be done later by
        # :meth:`Collector`
        if level not in Collector.raw_fixtures.keys():
            Collector.raw_fixtures[level] = []
        Collector.raw_fixtures[level].append(func)

        @functools.wraps(func)
        def wrapper_fixture(*args, **kwargs):
            func(*args, **kwargs)

        return wrapper_fixture
    return decorator_fixture



from __future__ import annotations

from _balder.connection import Connection
from _balder.decorator_insert_into_tree import insert_into_tree

from _balder.objects.connections import osi_3_network, osi_4_transport


@insert_into_tree(parents=[osi_4_transport.TcpIPv4Connection, osi_4_transport.TcpIPv6Connection])
class HttpConnection(Connection):
    """
    Balder HTTP connection (OSI LAYER 7)
    """


@insert_into_tree(parents=[osi_4_transport.TcpIPv4Connection, osi_4_transport.TcpIPv6Connection])
class ImapConnection(Connection):
    """
    Balder IMAP connection (OSI LAYER 7)
    """


@insert_into_tree(
    parents=[
        (osi_4_transport.TcpIPv4Connection, osi_4_transport.UdpIPv4Connection),
        (osi_4_transport.TcpIPv6Connection, osi_4_transport.UdpIPv6Connection),
        osi_3_network.IPv4Connection,
        osi_3_network.IPv6Connection
    ])
class LdapConnection(Connection):
    """
    Balder LDAP connection (OSI LAYER 7)
    """


@insert_into_tree(parents=[osi_4_transport.UdpIPv4Connection, osi_4_transport.UdpIPv6Connection])
class NtpConnection(Connection):
    """
    Balder NTP connection (OSI LAYER 7)
    """


@insert_into_tree(
    parents=[
        (osi_4_transport.UdpIPv4Connection, osi_4_transport.TcpIPv4Connection),
        (osi_4_transport.UdpIPv6Connection, osi_4_transport.TcpIPv6Connection)
    ])
class RpcConnection(Connection):
    """
    Balder RPC connection (OSI LAYER 7)
    """


@insert_into_tree(parents=[osi_4_transport.TcpIPv4Connection, osi_4_transport.TcpIPv6Connection])
class SmtpConnection(Connection):
    """
    Balder SMTP connection (OSI LAYER 7)
    """


@insert_into_tree(parents=[osi_4_transport.UdpIPv4Connection, osi_4_transport.UdpIPv6Connection])
class SntpConnection(Connection):
    """
    Balder SNTP connection (OSI LAYER 7)
    """


@insert_into_tree(parents=[osi_4_transport.TcpIPv4Connection, osi_4_transport.TcpIPv6Connection])
class SshConnection(Connection):
    """
    Balder SSH connection (OSI LAYER 7)
    """


@insert_into_tree(
    parents=[
        (osi_4_transport.UdpIPv4Connection, osi_4_transport.TcpIPv4Connection),
        (osi_4_transport.UdpIPv6Connection, osi_4_transport.TcpIPv6Connection)
    ])
class DnsConnection(Connection):
    """
    Balder DNS connection (OSI LAYER 7)
    """



import balder
from ..lib.features import AddCalculateFeature, ProvidesANumberFeature
from .scenario_adding import ScenarioAdding


class ScenarioAddingChild(ScenarioAdding):

    # remove all connection -> test should check if we use the parent connections in case there are no connections
    class Calculator(ScenarioAdding.Calculator):
        adds = AddCalculateFeature()

    # only set the connection for `NumberOneDevice`
    @balder.connect(Calculator, over_connection=balder.Connection)
    class NumberOneDevice(ScenarioAdding.NumberOneDevice):
        number = ProvidesANumberFeature()

    class NumberTwoDevice(ScenarioAdding.NumberTwoDevice):
        number = ProvidesANumberFeature()

    def test_add_two_numbers(self):
        self.NumberOneDevice.number.set_number(3)
        self.NumberOneDevice.number.sends_the_number()

        self.Calculator.adds.get_numbers()
        result = self.Calculator.adds.add_numbers()
        assert result == 3, "result is not as expected"



import balder
import balder.connections as cnns
from ..lib.features import AddCalculateFeature, ProvidesANumberFeature


class ScenarioAdding(balder.Scenario):

    class Calculator(balder.Device):
        # we could add vDevices here later too -> f.e. `NumberOne="NumberOneDevice", NumberTwo="NumberTwoDevice"`
        adds = AddCalculateFeature()

    # only set the connection for `NumberTwoDevice`
    class NumberOneDevice(balder.Device):
        number = ProvidesANumberFeature()

    @balder.connect(Calculator, over_connection=cnns.HttpConnection)
    class NumberTwoDevice(balder.Device):
        number = ProvidesANumberFeature()



from __future__ import annotations

from _balder.connection import Connection
from _balder.decorator_insert_into_tree import insert_into_tree

from _balder.objects.connections import osi_4_transport


@insert_into_tree(parents=[osi_4_transport.TcpIPv4Connection, osi_4_transport.TcpIPv6Connection])
class TelnetConnection(Connection):
    """
    Balder TELNET connection (OSI LAYER 6)
    """



from typing import Dict


def _compare_observed_entry(original_item: Dict, expected_item: Dict):
    """helper method that helps to compare the original with the expected entry"""
    KEY_FILE = "file"
    KEY_CLS = "cls"
    KEY_METH = "meth"
    KEY_MSG = "msg"
    KEY_CATEGORY = "category"
    KEY_PART = "part"

    ALL_POSSIBLE_KEYS = [KEY_FILE, KEY_CLS, KEY_METH, KEY_MSG, KEY_CATEGORY, KEY_PART]

    if KEY_FILE in expected_item.keys():
        if isinstance(expected_item[KEY_FILE], str) and \
                str(original_item[KEY_FILE]) != expected_item[KEY_FILE]:
            return False
    if KEY_CLS in expected_item.keys():
        if isinstance(expected_item[KEY_CLS], str) and original_item[KEY_CLS] is not None and \
                original_item[KEY_CLS] != expected_item[KEY_CLS]:
            return False
    if KEY_METH in expected_item.keys():
        if isinstance(expected_item[KEY_METH], str) and \
                original_item[KEY_METH] != expected_item[KEY_METH]:
            return False
    if KEY_MSG in expected_item.keys():
        if original_item[KEY_MSG] != expected_item[KEY_MSG]:
            return False
    if KEY_CATEGORY in expected_item.keys():
        if original_item[KEY_CATEGORY] != expected_item[KEY_CATEGORY]:
            return False
    if KEY_PART in expected_item.keys():
        if original_item[KEY_PART] != expected_item[KEY_PART]:
            return False
    for cur_key in expected_item.keys():
        if cur_key not in ALL_POSSIBLE_KEYS:
            raise KeyError(f"unknown key `{cur_key}` was given for expected_data entry dictionary")
    return True


def compare_observed_list_with_expected(observer_list, expected):
    """
    This function checks if the expected elements are given in the observer list. For this the method searches for the
    keys `file`, `cls`, `meth`, `msg`, `category` and `part`. It only checks the keys that are given in the
    dictionaries of the `expected_dict`.

    .. note::
        * LIST: means that the order of the elements does not matter
        * TUPLE: means that all elements in the tuple has to have the given order of the tuple
        * DICT: describes one object

        But note, that every element that exists in the `expected_tuple` has to be in the `observer_list`

    :param observer_list: the real observer list with the data

    :param expected: the expected data with information about the allowed order (List, Tuple, Dict)
    """
    # contains the index of the current observer item, that should be searched in `expected`
    cur_observer_idx = 0

    def go_through_tuple(of_tuple):
        """helps to go through a tuple -> returns False if there is no correct matching anymore otherwise it returns
        True (and updates the `cur_observer_idx`) after the whole tuple was checked"""
        nonlocal cur_observer_idx
        # in this case everything has to be in the exact order as the order given in the tuple
        for cur_tuple_item in of_tuple:
            if isinstance(cur_tuple_item, dict):
                # normal entry
                if _compare_observed_entry(original_item=observer_list[cur_observer_idx], expected_item=cur_tuple_item):
                    cur_observer_idx += 1
                else:
                    return False
            elif isinstance(cur_tuple_item, tuple):
                inner_result = go_through_tuple(cur_tuple_item)
                if not inner_result:
                    return False
            elif isinstance(cur_tuple_item, list):
                inner_result = go_through_list(cur_tuple_item)
                if not inner_result:
                    return False
            else:
                raise TypeError(f"unknown type `{type(cur_tuple_item)}`")
        return True

    def go_through_list(of_list):
        """helps to go through a list -> returns False if there is no correct matching anymore otherwise it returns
           True (and updates the `cur_observer_idx`) after the whole tuple was checked"""
        nonlocal cur_observer_idx
        leftover_items = of_list.copy()

        while len(leftover_items) > 0:
            length_before = len(leftover_items)
            idx_before = cur_observer_idx
            for cur_item in leftover_items:
                if isinstance(cur_item, dict):
                    # normal entry
                    if _compare_observed_entry(original_item=observer_list[cur_observer_idx],
                                               expected_item=cur_item):
                        cur_observer_idx += 1
                        leftover_items.remove(cur_item)
                        break
                elif isinstance(cur_item, tuple):
                    inner_result = go_through_tuple(cur_item)
                    if inner_result:
                        leftover_items.remove(cur_item)
                        break
                    else:
                        # revert index
                        cur_observer_idx = idx_before

                elif isinstance(cur_item, list):
                    inner_result = go_through_list(cur_item)
                    if inner_result:
                        leftover_items.remove(cur_item)
                        break
                    else:
                        # revert index
                        cur_observer_idx = idx_before
            if length_before == len(leftover_items):
                # can not resolve the whole list
                return False
        return True

    if isinstance(expected, tuple):
        result = go_through_tuple(expected)
    elif isinstance(expected, list):
        result = go_through_list(expected)
    else:
        raise TypeError(f"unknown type `{type(expected)}` given for the expected list")

    if not result:
        raise AssertionError(
            f"the execution was not as expected - problem to find the expected in the real data since index "
            f"`{cur_observer_idx}`")



import balder
from ...lib.features import FeatureII
from ...balderglob import RuntimeObserver
from .feature_i import SetupFeatureI


class SetupFeatureII(FeatureII):

    class OtherVDevice1(balder.VDevice):
        # this mechanism is not allowed - normally you can not map a vDevice inside a vDevice feature -> should throw
        # an error while collecting
        my_feature_i = SetupFeatureI(OtherVDevice2="OtherVDevice1")

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureII, SetupFeatureII.do_something, "enter `SetupFeatureII.do_something`",
            category="feature")



from __future__ import annotations
from typing import Type, Dict, Tuple, Union

from _balder.device import Device
from _balder.vdevice import VDevice


class Feature:
    """
    This is the basic feature class. It represents an abstract class that should not be used directly. It is the base
    class for all feature elements.
    """

    def __init__(self, **kwargs):
        """
        :param kwargs: contains a dictionary that references all vDevices of the feature and a real
                        scenario :meth:`Device` as value
        """
        from _balder.controllers import FeatureController

        #: this property contains the active mapping for the devices
        self.active_vdevices: Dict[VDevice, Union[Device, str]] = {}
        all_vdevices = FeatureController.get_for(self.__class__).get_abs_inner_vdevice_classes()
        not_processed_kwargs = {}
        for cur_kwargs_key, cur_kwargs_val in kwargs.items():
            if cur_kwargs_key in [cur_vdevice.__name__ for cur_vdevice in all_vdevices]:
                cur_vdevice = [cur_vdevice for cur_vdevice in all_vdevices if cur_vdevice.__name__ == cur_kwargs_key][0]
                if isinstance(cur_kwargs_val, str) or issubclass(cur_kwargs_val, Device):
                    if not isinstance(cur_kwargs_val, str) and issubclass(cur_kwargs_val, VDevice):
                        raise TypeError(f"the given value of vDevice mapping for vDevice `{cur_kwargs_key}` has to be "
                                        f"of the type `Device` - it is a subclass of `VDevice` - this is not allowed "
                                        f"here")
                    if self.active_vdevices:
                        raise AttributeError(
                            "the constructor expects exactly none or one vDevice mapping - found more than one here")

                    self.active_vdevices = {cur_vdevice: cur_kwargs_val}
                else:
                    raise TypeError(f"the given value of vDevice mapping for vDevice `{cur_kwargs_key}` has to be of "
                                    f"the type `Device` or has to be a string with the name of the device class")
            else:
                not_processed_kwargs[cur_kwargs_key] = cur_kwargs_val
        if len(not_processed_kwargs) != 0:
            raise TypeError(
                f"can not resolve the given attributes `{str(not_processed_kwargs)}` in feature "
                f"`{self.__class__.__name__}`")

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS ----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def active_vdevice_device_mapping(self) -> Tuple[Union[None, Type[VDevice]], Union[None, Type[Device]]]:
        """This property returns the mapped device that is active here"""
        if len(self.active_vdevices) == 0:
            return None, None
        return list(self.active_vdevices.keys())[0], list(self.active_vdevices.values())[0]

    @property
    def active_vdevice(self) -> Union[None, Type[VDevice]]:
        """This property returns the active VDevice that is active here"""
        return self.active_vdevice_device_mapping[0]

    @property
    def active_mapped_device(self) -> Union[None, Type[VDevice]]:
        """This property returns the active mapped Device"""
        return self.active_vdevice_device_mapping[1]

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    # ---------------------------------- METHODS -----------------------------------------------------------------------
